import warnings
import pandas as pd
import numpy as np
from sklearn import tree, ensemble, neighbors, linear_model, neural_network, svm, naive_bayes
from sklearn import preprocessing, model_selection, exceptions
from PyQt5.QtWidgets import QTreeWidgetItem
from matplotlib.backends.backend_qt5agg import FigureCanvas
from matplotlib.figure import Figure
from matplotlib import pyplot as plt
from matplotlib import rcParams
from matplotlib.colors import ListedColormap

from ml4bio.model_metrics import ModelMetrics
# from model_metrics import ModelMetrics

class Model:
	"""
	An instance of this class stores information about a trained classifier.
	In particular, it contains:

		- data for training and evaluation
		- classifier object
		- classifier's hyperparameters
		- classifier's name
		- user's comment on the classifier
		- performance metrics on training data
		- performance metrics on validation data
		- performance metrics on test data (if the classifier has been tested)

	:cvar counter: number of existing classifiers, used in default classifier names.
	:vartype counter: int
	:ivar classifier: the classifier object, see sklearn documentation for details.
	:vartype classifier: sklearn classifier object
	:ivar X: feature values of training data (including training and validation sets)
	:vartype X: pandas dataframe
	:ivar y: labels of training data
	:vartype y: pandas series
	:ivar name\_: name of the classifier
	:vartype name\_: str
	:ivar type\_: type of the classifier (defined in child classes)
	:vartype type\_: str
	:ivar comment\_: user comment on the classifier
	:vartype comment\_: str
	:ivar params\_: hyperparameters of the classifier (see sklearn classifier documentation for details)
	:vartype params\_: dict
	:ivar train_metrics: performance metrics of the classifier on training data
	:vartype train_metrics: ModelMetrics
	:ivar val_metrics: performance metrics of the classifier on validation data
	:vartype val_metrics: ModelMetrics
	:ivar test_metrics: performance metrics of the classifier on test data
	:vartype test_metrics: ModelMetrics
	"""

	counter = 0 	# counts the number of trained classifiers.

	def __init__(self, classifier, X, y, val_method, val_size, k, stratify):
		"""
		Constructs a new instance of the class.

		:param classifier: the classifier object, see sklearn documentation for details.
		:type classifier: sklearn classifier object
		:param X: feature values of training data (including training and validation sets)
		:type X: pandas dataframe
		:param y: labels of training data
		:type y: pandas series
		:param val_method: validation method ('holdout', 'cv', 'loo')
		:type val_method: str
		:param val_size: percent of validation data (only used when val_method='holdout')
		:type val_size: float
		:param k: number of folds (only used when val_method='cv')
		:type k: int
		:param stratify: draw samples according to class proportions or not
		:type stratify: bool
		"""
		Model.counter += 1

		self.classifier = classifier
		self.X = X
		self.y = y

		# default name for the classifier
		self.name_ = 'classifier_' + str(Model.counter)
		self.type_ = ''
		self.comment_ = ''
		self.params_ = classifier.get_params()	# hyperparameters of classifier

		if val_method == 'holdout':
			self.train_metrics, self.val_metrics = self.__hold_out_validation(\
				classifier, X, y, val_size, stratify)
		elif val_method == 'cv':
			self.train_metrics, self.val_metrics = self.__cross_validation(\
				classifier, X, y, k, stratify)
		elif val_method == 'loo':
			self.train_metrics, self.val_metrics = self.__cross_validation(\
				classifier, X, y, X.shape[0])

		self.test_metrics = None

	def __hold_out_validation(self, classifier, X, y, val_size, stratify=True):
		"""
		Performs classifier validation using held-out data.

		:param classifier: classifier for validation
		:type classifier: sklearn classifier object
		:param X: feature values of training data (including training and validation sets)
		:type X: pandas dataframe
		:param y: labels of training data
		:type y: pandas series
		:param val_size: percent of validation data
		:type val_size: float
		:param stratify: draw samples according to class proportions or not
		:type stratify: bool

		:returns: performance metrics on training and validation data
		"""
		if not stratify or val_size == 0:
			s = None
		else:
			s = y

		X_train, X_val, y_train, y_val = model_selection.train_test_split(\
			X, y, test_size=val_size, stratify=s, random_state=0)

		# catch convergence warning
		with warnings.catch_warnings():
			warnings.filterwarnings('error', category=exceptions.ConvergenceWarning)
			try:
				classifier = classifier.fit(X_train, y_train)
			except exceptions.ConvergenceWarning:
				Model.counter -= 1
				raise

		y_train_pred = classifier.predict(X_train) 			# class prediction
		y_train_prob = classifier.predict_proba(X_train)	# probability of each class

		if X_val.shape[0] != 0:
			y_val_pred = classifier.predict(X_val)
			y_val_prob = classifier.predict_proba(X_val)

			return ModelMetrics(classifier, y_train, y_train_pred, y_train_prob, 'holdout'), \
				ModelMetrics(classifier, y_val, y_val_pred, y_val_prob, 'holdout')
		else:
			return ModelMetrics(classifier, y_train, y_train_pred, y_train_prob, 'holdout'), None

	def __cross_validation(self, classifier, X, y, k, stratify=True):
		"""
		Performs classifier validation through cross-validation.
		This function is also used by leave-one-out validation.

		:param classifier: classifier for validation
		:type classifier: sklearn classifier object
		:param X: feature values of training data (including training and validation sets)
		:type X: pandas dataframe
		:param y: labels of training data
		:type y: pandas series
		:param k: number of folds
		:type k: int
		:param stratify: draw samples according to class proportions or not
		:type stratify: bool

		:returns: performance metrics on training and validation data
		"""
		if k == X.shape[0]:		# leave-one-out
			kf = model_selection.KFold(n_splits=k)
		else:
			if stratify:
				kf = model_selection.StratifiedKFold(n_splits=k, shuffle=True, random_state=0)
			else:
				kf = model_selection.KFold(n_splits=k, shuffle=True, random_state=0)

		# training data and predictions for each fold
		y_train_list = []
		y_train_pred_list = []
		y_train_prob_list = []
		y_val_list = []
		y_val_pred_list = []
		y_val_prob_list = []

		for train_idx, val_idx in kf.split(X, y):
			X_train, X_val = X.iloc[train_idx], X.iloc[val_idx]
			y_train, y_val = y.iloc[train_idx], y.iloc[val_idx]
			y_train_list.append(y_train)
			y_val_list.append(y_val)

			# catch convergence warning
			with warnings.catch_warnings():
				warnings.filterwarnings('error', category=exceptions.ConvergenceWarning)
				try:
					classifier = classifier.fit(X_train, y_train)
				except exceptions.ConvergenceWarning:
					Model.counter -= 1
					raise

			y_train_pred_list.append(classifier.predict(X_train))
			y_val_pred_list.append(classifier.predict(X_val))
			y_train_prob_list.append(classifier.predict_proba(X_train))
			y_val_prob_list.append(classifier.predict_proba(X_val))

		if k == X.shape[0]:		# leave-one-out
			y_val = np.hstack(y_val_list)
			y_val_pred = np.hstack(y_val_pred_list)
			y_val_prob = np.vstack(y_val_prob_list)

			return ModelMetrics(classifier, y_train_list, y_train_pred_list, y_train_prob_list, 'cv'), \
				ModelMetrics(classifier, y_val, y_val_pred, y_val_prob, 'loo')
		else:
			return ModelMetrics(classifier, y_train_list, y_train_pred_list, y_train_prob_list, 'cv'), \
				ModelMetrics(classifier, y_val_list, y_val_pred_list, y_val_prob_list, 'cv')

	def test(self, X, y):
		"""
		Tests a classifier.
		This function adds performance metrics on test data.

		:param X: feature values for test data
		:type X: pandas dataframe
		:param y: labels for test data
		:type y: pandas series
		"""
		self.test_X = X
		self.test_y = y

		classifier = self.classifier.fit(self.X, self.y)
		y_pred = classifier.predict(X) 			# class prediction
		y_prob = classifier.predict_proba(X)	# probability of each class
		self.test_metrics =  ModelMetrics(classifier, y, y_pred, y_prob, 'holdout')

	def predict(self, X, path):
		"""
		Predicts labels for new data using the classifier.
		Saves the prediction in a new file.

		:param X: features values of new data
		:type X: pandas dataframe
		:param path: where the prediction is saved
		:type path: str
		"""
		classifier = self.classifier.fit(self.X, self.y)
		y_pred = pd.DataFrame(classifier.predict(X), columns=['prediction'])
		y_prob = pd.DataFrame(np.around(classifier.predict_proba(X), decimals=4), columns=classifier.classes_)
		output = pd.concat([X, y_pred, y_prob], axis=1)
		output.to_csv(path, sep=',', index=False) 		# save to file

	def set_name(self, name):
		"""
		Sets name of the classifier.

		:param name: user-supplied classifier name
		:type name: str
		"""
		self.name_ = name

	def set_comment(self, comment):
		"""
		Sets comment on the classifier.

		:param comment: user-supplied comment
		:type comment: str
		"""
		self.comment_ = comment

	def name(self):
		"""
		Returns name of the classifier.

		:returns: name of the classifier
		"""
		return self.name_

	def type(self):
		"""
		Returns type of the classifier.

		:returns: type of the classifier
		"""
		return self.type_

	def comment(self):
		"""
		Returns comment on the classifier.

		:returns: comment on the classifier
		"""
		return self.comment_

	def params(self):
		"""
		Returns hyperparameters of the classifier.

		:returns: hyperparameters of the classifier
		"""
		return self.params_

	def metrics(self, option):
		"""
		Returns performance metrics on the specified data.

			- option='train': metrics on training data
			- option='val': metrics on validation data
			- option='test': metrics on test data

		:param option: 'train', 'val' or 'test'
		:type option: str

		:returns: performance metrics on the specified data
		"""
		if option == 'train':	return self.train_metrics
		elif option == 'val':	return self.val_metrics
		elif option == 'test':	return self.test_metrics

	def summary(self, view):
		"""
		Displays a summary of classifier performance. 
		The summary is organized in a tree structure.

		:param view: the widget where the summary is displayed
		:type view: QTreeWidget
		"""
		view.clear()

		name_item = QTreeWidgetItem(view)
		name_item.setText(0, 'Name')
		name_item.setText(1, self.name())
		name_item.setToolTip(1, self.name())
		type_item = QTreeWidgetItem(view)
		type_item.setText(0, 'Type')
		type_item.setText(1, self.type())
		params_item = QTreeWidgetItem(view)
		params_item.setText(0, 'Parameters')
		
		# show classifier hyperparameters
		params = self.params()
		for param in params:
			param_item = QTreeWidgetItem()
			param_item.setText(0, param)
			param_item.setToolTip(0, param)
			value = params[param]
			if value == None:
				param_item.setText(1, 'None')
			else:
				param_item.setText(1, str(value))
			params_item.addChild(param_item)

		metrics_item = QTreeWidgetItem(view)
		metrics_item.setText(0, 'Performance')

		# show metrics on training data
		train_metrics_item = QTreeWidgetItem(metrics_item)
		train_metrics_item.setText(0, 'Training')
		train_metrics = self.metrics('train')
		train_metrics.summary(train_metrics_item)

		# show metrics on validation data
		val_metrics_item = QTreeWidgetItem(metrics_item)
		val_metrics_item.setText(0, 'Validation')
		val_metrics = self.metrics('val')
		if val_metrics is not None:
			val_metrics.summary(val_metrics_item)

		# show metrics on test data
		test_metrics = self.metrics('test')
		if test_metrics is not None:
			test_metrics_item = QTreeWidgetItem(metrics_item)
			test_metrics_item.setText(0, 'Test')
			test_metrics.summary(test_metrics_item)

		comment_item = QTreeWidgetItem(view)
		comment_item.setText(0, 'Comment')
		comment_text_item = QTreeWidgetItem(comment_item)
		comment_text_item.setFirstColumnSpanned(True)
		comment_text_item.setText(0, self.comment())

	def plot_decision_regions(self, option, canvas):
		"""
		Plots data with decision regions.
		This is only enabled for data with 2D continuous features.

		Reference:
			http://scikit-learn.org/stable/auto_examples/classification/plot_classifier_comparison.html

		:param option: 'train' or 'test'
		:type option: pandas dataframe
		:param canvas: the canvas where the plot is drawn
		:type canvas: FigureCanva
		"""
		le = preprocessing.LabelEncoder()		# integer encoder
		le.fit(self.y)
		classifier = self.classifier.fit(self.X, le.transform(self.y))
		classes = classifier.classes_
		num_classes = len(classes)

		if option == 'train':
			X = self.X
			y = self.y
		elif option == 'test':
			X = self.test_X
			y = self.test_y

		b1 = self.X.iloc[:, 0]
		b2 = self.X.iloc[:, 1]
		b1_slack = (b1.max() - b1.min()) * 0.1
		b2_slack = (b2.max() - b2.min()) * 0.1
		b1_min, b1_max = b1.min() - b1_slack, b1.max() + b1_slack 	# x-axis range
		b2_min, b2_max = b2.min() - b2_slack, b2.max() + b2_slack	# y-axis range
		step_1 = (b1_max - b1_min) / 200
		step_2 = (b2_max - b2_min) / 200
		md1, md2 = np.meshgrid(np.arange(b1_min, b1_max, step_1), np.arange(b2_min, b2_max, step_2))

		rcParams.update({'font.size': 7})
		canvas.figure.clear()
		ax = canvas.figure.subplots()
		levels = np.arange(-0.19, 1, 0.2) + 0.2

		if num_classes == 2:
			cm_bkgd = plt.cm.RdBu
			cm_pts = ListedColormap(['#FF0000', '#0000FF'])
			Z = classifier.predict_proba(np.c_[md1.ravel(), md2.ravel()])[:, 1]
			Z = Z.reshape(md1.shape)
			ax.contourf(md1, md2, Z, vmin=0, vmax=1, cmap=cm_bkgd, alpha=0.8)
		
		elif num_classes == 3:
			cm_bkgd_1 = plt.cm.Reds
			cm_bkgd_2 = plt.cm.Greens
			cm_bkgd_3 = plt.cm.Blues
			cm_pts = cm_pts = ListedColormap(['#FF0000', '#00FF00', '#0000FF'])
			Z = classifier.predict_proba(np.c_[md1.ravel(), md2.ravel()])
			Z1 = Z[:, 0]
			Z2 = Z[:, 1]
			Z3 = Z[:, 2]

			P1 = np.maximum(0, Z1 - np.maximum(Z2, Z3))
			P2 = np.maximum(0, Z2 - np.maximum(Z1, Z3))
			P3 = np.maximum(0, Z3 - np.maximum(Z1, Z2))
			P1 = P1.reshape(md1.shape)
			P2 = P2.reshape(md1.shape)
			P3 = P3.reshape(md1.shape)

			ax.contourf(md1, md2, P1, levels, cmap=cm_bkgd_1, alpha=0.8)
			ax.contourf(md1, md2, P2, levels, cmap=cm_bkgd_2, alpha=0.8)
			ax.contourf(md1, md2, P3, levels, cmap=cm_bkgd_3, alpha=0.8)

		d1 = X.iloc[:, 0] 	# x-axis
		d2 = X.iloc[:, 1]	# y-axis
		ax.scatter(d1, d2, c=le.transform(y), cmap=cm_pts, alpha=0.6, edgecolors='k')
		ax.set_xlim(md1.min(), md1.max())
		ax.set_ylim(md2.min(), md2.max())
		ax.set_xticks(())
		ax.set_yticks(())

		canvas.figure.tight_layout()
		canvas.draw()

	def clear():
		"""
		Starts the counter from zero.
		"""
		Model.counter = 0

class DecisionTree(Model):
	"""
	An instance of this class stores information about a trained decision tree classifier.
	It inherits the **Model** class.
	"""
	def __init__(self, classifier, X, y, val_method, val_size, k, stratify):
		"""
		See **Model** class.
		"""
		super().__init__(classifier, X, y, val_method, val_size, k, stratify)		
		self.type_ = 'decision tree'

class RandomForest(Model):
	"""
	An instance of this class stores information about a trained random forest classifier.
	It inherits the **Model** class.
	"""
	def __init__(self, classifier, X, y, val_method, val_size, k, stratify):
		"""
		See **Model** class.
		"""
		super().__init__(classifier, X, y, val_method, val_size, k, stratify)
		self.type_ = 'random forest'

class KNearestNeighbors(Model):
	"""
	An instance of this class stores information about a trained K-nearest neighbors classifier.
	It inherits the **Model** class.
	"""
	def __init__(self, classifier, X, y, val_method, val_size, k, stratify):
		"""
		See **Model** class.
		"""
		super().__init__(classifier, X, y, val_method, val_size, k, stratify)
		self.type_ = 'k-nearest neighbors'

class LogisticRegression(Model):
	"""
	An instance of this class stores information about a trained logistic regression classifier.
	It inherits the **Model** class.
	"""
	def __init__(self, classifier, X, y, val_method, val_size, k, stratify):
		"""
		See **Model** class.
		"""
		super().__init__(classifier, X, y, val_method, val_size, k, stratify)
		self.type_ = 'logistic regression'

class NeuralNetwork(Model):
	"""
	An instance of this class stores information about a trained neural network classifier.
	It inherits the **Model** class.
	"""
	def __init__(self, classifier, X, y, val_method, val_size, k, stratify):
		"""
		See **Model** class.
		"""
		super().__init__(classifier, X, y, val_method, val_size, k, stratify)
		self.type_ = 'neural network'

class SVM(Model):
	"""
	An instance of this class stores information about a trained SVM classifier.
	It inherits the **Model** class.
	"""
	def __init__(self, classifier, X, y, val_method, val_size, k, stratify):
		"""
		See **Model** class.
		"""
		super().__init__(classifier, X, y, val_method, val_size, k, stratify)
		self.type_ = 'svm'

class NaiveBayes(Model):
	"""
	An instance of this class stores information about a trained naive bayes classifier.
	It inherits the **Model** class.
	"""
	def __init__(self, classifier, X, y, val_method, val_size, k, stratify, mode):
		"""
		See **Model** class. The extra parameter *mode* indicates gaussian or multinomial NB.

		:param mode: 'gaussian' or 'multinomial'
		:type mode: str
		"""
		super().__init__(classifier, X, y, val_method, val_size, k, stratify)
		self.type_ = 'naive bayes'


import os
import pandas as pd
import numpy as np
from sklearn import preprocessing, model_selection
from PyQt5.QtWidgets import QTreeWidgetItem

class Data:
	"""
	An instance of this class stores the data used for learning and prediction.
	The stored information includes:

		- labeled data
		- unlabeled data (if uploaded)
		- labeled file name
		- unlabeled file name (if uploaded)
		- number of samples in labeled data
		- number of samples in unlabeled data (if uploaded)
		- number of features
		- type of features (individual and overall)
		- number of valid classes
		- number of samples per class
		- encoded data (integer and one-hot encoding)

	:ivar labeled_data: labeled data for learning classifiers
	:vartype labeled_data: pandas dataframe
	:ivar unlabeled_data: unlabeled data for prediction
	:vartype unlabeled_data: pandas dataframe
	:ivar labeled_name: name of the labeled dataset
	:vartype labeled_name: str
	:ivar unlabeled_name: name of the unlabeled dataset
	:vartype unlabeled_name: str
	:ivar labeled_num_samples: number of samples in the labeled dataset
	:vartype labeled_num_samples: int
	:ivar unlabeled_num_samples: number of samples in the unlabeled dataset
	:vartype unlabeled_num_samples: int
	:ivar num_features\_: number of features
	:vartype num_features\_: int
	:ivar individual_feature_type: type of each feature in a dictionary whose keys are feature names
	:vartype individual_feature_type: dict
	:ivar global_feature_type: an overall summary of feature types ('continuous', 'discrete', 'mixed')
	:vartype global_feature_type: str
	:ivar num_classes\_: number of valid classes
	:vartype num_classes\_: int
	:ivar class_num_samples: number of samples per class in a dictionary whose keys are class names
	:vartype class_num_samples: dict
	:ivar integer_encoded_labeled_data: integer-encoded labeled data
	:vartype integer_encoded_labeled_data: pandas dataframe
	:ivar integer_encoded_unlabeled_data: integer-encoded unlabeled data
	:vartype integer_encoded_unlabeled_data: pandas dataframe
	:ivar one_hot_encoded_labeled_data: one-hot-encoded labeled data
	:vartype one_hot_encoded_labeled_data: pandas dataframe
	:ivar one_hot_encoded_unlabeled_data: one-hot-encoded unlabeled data
	:vartype one_hot_encoded_unlabeled_data: pandas dataframe
	:ivar num_one_hot_encoded_features: number of features after one-hot encoding
	:vartype num_one_hot_encoded_features: int
	"""

	def __init__(self, path):
		"""
		Constructs a new instance of the class.

		:param path: file path of the labeled data
		:type path: str
		"""
		self.labeled_data = pd.read_csv(path, sep=None, engine='python')
		self.unlabeled_data = None 		# unlabeled data is optional
		self.labeled_name = os.path.basename(path)
		self.labeled_num_samples = self.labeled_data.shape[0]

		self.num_features_ = self.labeled_data.shape[1] - 1
		feature_names = self.labeled_data.columns[0: self.num_features_]
		num_continuous_features = 0
		num_discrete_features = 0
		
		# determine feature types
		self.individual_feature_type = dict()
		for f in feature_names:
			dtype = self.labeled_data.dtypes[f]
			if pd.api.types.is_string_dtype(dtype):
				num_discrete_features += 1
				self.individual_feature_type[f] = 'discrete'
			elif pd.api.types.is_integer_dtype(dtype):
				if len(set(self.labeled_data[f])) <= 10:
					num_discrete_features += 1
					self.individual_feature_type[f] = 'discrete'
				else:
					num_continuous_features += 1
					self.individual_feature_type[f] = 'continuous'
			else:
				num_continuous_features += 1
				self.individual_feature_type[f] = 'continuous'

		# determine global feature type
		if num_continuous_features > 0 and num_discrete_features == 0:
			self.global_feature_type = 'continuous'
		elif num_continuous_features == 0 and num_discrete_features > 0:
			self.global_feature_type = 'discrete'
		else:
			self.global_feature_type = 'mixed'

		label_col = self.labeled_data.iloc[:, self.num_features_]
		classes = list(set(label_col))
		self.num_classes_ = len(classes)
		
		# count number of samples that belong to each class
		self.class_num_samples = dict()
		for c in classes:
			self.class_num_samples[c] = 0
		for l in label_col:
			self.class_num_samples[l] += 1

	def add_unlabeled_data(self, path):
		"""
		Adds unlabeled data for prediction.

		:param path: file path of the unlabeled data
		:type path: str
		"""
		try:
			self.unlabeled_data = pd.read_csv(path, sep=None, engine='python')
		except:
			raise
		
		labeled_feature_names = list(self.labeled_data.columns[0: self.num_features()])
		unlabeled_feature_names = list(self.unlabeled_data.columns)
		if labeled_feature_names != unlabeled_feature_names:
			self.unlabeled_data = None
			raise ValueError()
		
		self.unlabeled_name = os.path.basename(path)
		self.unlabeled_num_samples = self.unlabeled_data.shape[0]

	def encode(self):
		"""
		Encodes the data using integer and one-hot encoders.
		"""
		raw_data = self.labeled_data.iloc[:, 0: self.num_features()]
		if self.unlabeled_data is not None:
			raw_data = pd.concat([raw_data, self.unlabeled_data], axis=0)

		integer_encoded_data = pd.DataFrame([])
		one_hot_encoded_data = pd.DataFrame([])
		le = preprocessing.LabelEncoder()		# integer encoder
		ohe = preprocessing.OneHotEncoder()		# one-hot encoder

		feature_type = self.feature_type(option='individual')
		for i in range(0, self.num_features()):
			feature = raw_data.columns[i]
			if feature_type[feature] == 'discrete':
				le_col = le.fit_transform(raw_data[feature])
				num_values = len(np.unique(le_col))	# number of values of a discrete feature
				le_col = pd.DataFrame(le_col)
				le_col.columns = [feature]
				integer_encoded_data = pd.concat([integer_encoded_data, le_col], axis=1)

				ohe_cols = pd.DataFrame(ohe.fit_transform(pd.DataFrame(le_col)).toarray())
				values = list(le.inverse_transform(list(range(0, num_values))))	# a collection of values
				col_names = [feature + '=' + str(v) for v in values]	# construct descriptive column names
				ohe_cols.columns = col_names
				one_hot_encoded_data = pd.concat([one_hot_encoded_data, ohe_cols], axis=1)
			else:
				integer_encoded_data = pd.concat([integer_encoded_data, raw_data[feature]], axis=1)
				one_hot_encoded_data = pd.concat([one_hot_encoded_data, raw_data[feature]], axis=1)

		self.integer_encoded_labeled_data = integer_encoded_data[0: self.num_samples('labeled')] 	# no label column
		self.one_hot_encoded_labeled_data = one_hot_encoded_data[0: self.num_samples('labeled')]	# no label column
		self.num_one_hot_encoded_features = len(self.one_hot_encoded_labeled_data.columns)

		if self.unlabeled_data is not None:
			self.integer_encoded_unlabeled_data = integer_encoded_data[self.num_samples('labeled'): ]
			self.one_hot_encoded_unlabeled_data = one_hot_encoded_data[self.num_samples('labeled'): ]

	def split(self, test_size, stratify):
		"""
		Splits labeled data into training and test sets.

		:param test_size: percent of test data
		:type test_size: float
		:param stratify: draw samples according to class proportions or not
		:type stratify: bool
		"""
		integer_encoded_X = self.integer_encoded_labeled_data
		one_hot_encoded_X = self.one_hot_encoded_labeled_data
		y = self.labeled_data.iloc[:, self.num_features()]
		
		if stratify:
			s = y
		else:
			s = None

		integer_encoded_X_train, integer_encoded_X_test, integer_encoded_y_train, integer_encoded_y_test = \
			model_selection.train_test_split(integer_encoded_X, y, test_size=test_size, stratify=s, random_state=0)
		one_hot_encoded_X_train, one_hot_encoded_X_test, one_hot_encoded_y_train, one_hot_encoded_y_test = \
			model_selection.train_test_split(one_hot_encoded_X, y, test_size=test_size, stratify=s, random_state=0)

		self.integer_encoded_train = pd.concat([integer_encoded_X_train, integer_encoded_y_train], axis=1)
		self.integer_encoded_test = pd.concat([integer_encoded_X_test, integer_encoded_y_test], axis=1)
		self.one_hot_encoded_train = pd.concat([one_hot_encoded_X_train, one_hot_encoded_y_train], axis=1)
		self.one_hot_encoded_test = pd.concat([one_hot_encoded_X_test, one_hot_encoded_y_test], axis=1)

	def name(self, option='labeled'):
		"""
		Returns name of the dataset.

			- option='labeled': name of labeled data
			- option='unlabeled': name of unlabeled data

		:param option: 'labeled' or 'unlabeled'
		:type option: str

		:returns: name of the dataset
		"""
		if option == 'labeled':		return self.labeled_name
		elif option == 'unlabeled':	return self.unlabeled_name

	def num_samples(self, option='labeled'):
		"""
		Returns number of samples.

			- option='labeled': total number of samples in labeled data
			- option='unlabeled': total number of samples in unlabeled data
			- option='classwise': number of samples that belong to each class

		:param option: 'labeled', 'unlabeled' or 'classwise'
		:type option: str

		:returns: number of samples
		"""
		if option == 'labeled':		return self.labeled_num_samples
		elif option == 'unlabeled':	return self.unlabeled_num_samples
		elif option == 'classwise':	return self.class_num_samples

	def num_features(self, option='raw'):
		"""
		Returns number of features.

			- option='raw': number of raw features (raw and integer-encoded data)
			- option='one-hot': number of features after one-hot encoding

		:param option: 'raw','integer' or 'one-hot'
		:type option: str

		:returns: number of features
		"""
		if option in ['raw', 'integer']:	return self.num_features_
		elif option == 'one-hot':			return self.num_one_hot_encoded_features

	def feature_type(self, option='global'):
		"""
		Returns summary of feature types.

			- option='individual': type of each feature
			- option='global': overall summary of feature types

		:param option: 'individual' or 'global'
		:type option: str

		:returns: summary of feature types
		"""
		if option == 'individual':	return self.individual_feature_type
		elif option == 'global':	return self.global_feature_type

	def num_classes(self):
		"""
		Returns number of valid classes.

		:returns: number of valid classes
		"""
		return self.num_classes_

	def summary(self, view):
		"""
		Displays a summary of the loaded data. This includes 

			- number of samples (total and classwise)
			- number of features
			- feature types (individual and global)

		The summary is organized in a tree structure.

		:param view: the widget where the summary is displayed
		:type view: QTreeWidget
		"""
		view.clear()

		samples_item = QTreeWidgetItem(view)
		samples_item.setText(0, 'Samples')
		labeled_samples_item = QTreeWidgetItem(samples_item)
		labeled_samples_item.setText(0, 'labeled')
		labeled_samples_item.setText(1, str(self.num_samples()))
		
		# show number of samples that belong to each class
		class_num_samples = self.num_samples('classwise')
		for c in class_num_samples:
			class_samples_item = QTreeWidgetItem(labeled_samples_item)
			class_samples_item.setText(0, str(c))
			class_samples_item.setText(1, str(class_num_samples[c]))
			class_samples_item.setToolTip(0, str(c))
		
		if self.unlabeled_data is not None:
			unlabeled_samples_item = QTreeWidgetItem(samples_item)
			unlabeled_samples_item.setText(0, 'unlabeled')
			unlabeled_samples_item.setText(1, str(self.num_samples('unlabeled')))

		features_item = QTreeWidgetItem(view)
		features_item.setText(0, 'Features')
		features_item.setText(1, str(self.num_features()))
		
		# show type of each feature
		feature_type = self.feature_type(option='individual')
		for f in feature_type:
			feature_type_item = QTreeWidgetItem(features_item)
			feature_type_item.setText(0, f)
			feature_type_item.setText(1, feature_type[f])
			feature_type_item.setToolTip(0, f)

	def train(self, option):
		"""
		Returns encoded training data.

			- option='integer': integer-encoded data
			- option='one-hot': one-hot-encoded data

		:param option: 'integer' or 'one-hot'
		:type option: str

		:returns: encoded training data
		"""
		if option == 'integer':		return self.integer_encoded_train
		elif option == 'one-hot':	return self.one_hot_encoded_train

	def test(self, option):
		"""
		Returns encoded test data.

			- option='integer': integer-encoded data
			- option='one-hot': one-hot-encoded data	

		:param option: 'integer' or 'one-hot'
		:type option: str

		:returns: encoded test data
		"""
		if option == 'integer':		return self.integer_encoded_test
		elif option == 'one-hot':	return self.one_hot_encoded_test

	def prediction(self, option):
		"""
		Returns encoded unlabeled data for prediction.

			- option='integer': integer-encoded data
			- option='one-hot': one-hot-encoded data

		:param option: 'integer' or 'one-hot'
		:type option: str

		:returns: encoded unlabeled data
		"""
		if option == 'integer':		return self.integer_encoded_unlabeled_data
		elif option == 'one-hot':	return self.one_hot_encoded_unlabeled_data


import os, sys, warnings, webbrowser
import pandas as pd
from pandas import errors
import numpy as np
from sklearn import tree, ensemble, neighbors, linear_model, neural_network, svm, naive_bayes, exceptions
from matplotlib.backends.backend_qt5agg import FigureCanvas
from matplotlib.figure import Figure
from PyQt5 import QtCore
from PyQt5.QtCore import QSize, Qt, QThread, pyqtSignal
from PyQt5.QtWidgets import QMainWindow, QApplication, QWidget, QFileDialog, QMessageBox
from PyQt5.QtWidgets import QPushButton, QRadioButton
from PyQt5.QtWidgets import QComboBox, QSpinBox, QDoubleSpinBox, QCheckBox, QLineEdit, QTextEdit, QLabel
from PyQt5.QtWidgets import QStackedWidget, QGroupBox, QFrame, QTableWidget, QTreeWidget, QTableWidgetItem, QTreeWidgetItem, QListView
from PyQt5.QtWidgets import QFormLayout, QGridLayout, QHBoxLayout, QVBoxLayout, QSpacerItem, QSizePolicy
from PyQt5.QtGui import QFont, QIcon, QPixmap

from ml4bio.data import Data
from ml4bio.model import Model, DecisionTree, RandomForest, KNearestNeighbors, LogisticRegression, NeuralNetwork, SVM, NaiveBayes
from ml4bio.model_metrics import ModelMetrics

# from data import Data
# from model import Model, DecisionTree, RandomForest, KNearestNeighbors, LogisticRegression, NeuralNetwork, SVM, NaiveBayes
# from model_metrics import ModelMetrics

class Training_thread(QThread):
    """
    An instance of this class is a thread for training classifiers.
    It is different than the main thread in which the main window is run.
    By creating a new thread for training, the main window is not frozen.
    The user may inspect existing classifiers when training is in progress.
    This class inherits **QThread**.

    :ivar app: main window of the software (an instance of class *App*)
    :vartype app: App
    """

    finished = pyqtSignal(object)   # SIGNAL: training done
    error = pyqtSignal(str)         # SIGNAL: show error message
    info = pyqtSignal(str)          # SIGNAL: show info message

    def __init__(self, app):
        """
        Constructs a new thread for training.

        :param app: main window of the software (an instance of class *App*)
        :type app: App
        """
        super().__init__()
        self.app = app

    def __del__(self):
        self.wait()

    def _train(self):
        """
        Trains a classifier and returns the classifier to the main thread.
        """
        index = self.app.paramStack.currentIndex()
        if index == 1:      model = self.decision_tree()
        elif index == 2:    model = self.random_forest()
        elif index == 3:    model = self.k_nearest_neighbors()
        elif index == 4:    model = self.logistic_regression()
        elif index == 5:    model = self.neural_network()
        elif index == 6:    model = self.svm()
        elif index == 7:    model = self.naive_bayes()
        
        self.finished.emit(model)   # passes the classifier to the main thread.

    def run(self):
        """
        Defines the behavior of the new thread once created.
        Here, training starts immediately in the new thread.
        """
        self._train()

    def decision_tree(self):
        """
        Trains and returns a decision tree classifier.
        """
        data = self.app.data.train('integer')
        num_features = self.app.data.num_features('integer')
        X = data.iloc[:, 0: num_features]
        y = data.iloc[:, num_features]

        max_depth = self.app.dtMaxDepthLineEdit.text()
        if max_depth == 'None':
            max_depth = None
        else:
            try:
                max_depth = int(max_depth)
            except:
                self.error.emit('max_depth')
                return
            
            if max_depth <= 0 or float(self.app.dtMaxDepthLineEdit.text()) - max_depth != 0:
                self.error.emit('max_depth')
                return

        class_weight = self.app.dtClassWeightComboBox.currentText()
        if class_weight == 'uniform':
            class_weight = None

        dt = tree.DecisionTreeClassifier(\
            criterion = self.app.dtCriterionComboBox.currentText(), \
            max_depth = max_depth, \
            min_samples_split = self.app.dtMinSamplesSplitSpinBox.value(), \
            min_samples_leaf = self.app.dtMinSamplesLeafSpinBox.value(), \
            class_weight = class_weight, \
            random_state = 0)

        return DecisionTree(dt, X, y, val_method=self.app.val_method, \
                val_size=self.app.holdoutSpinBox.value() / 100, \
                k = self.app.cvSpinBox.value(), \
                stratify=self.app.validationCheckBox.isChecked())

    def random_forest(self):
        """
        Trains and returns a random forest classifier.
        """
        data = self.app.data.train('integer')
        num_features = self.app.data.num_features('integer')
        X = data.iloc[:, 0: num_features]
        y = data.iloc[:, num_features]

        max_depth = self.app.rfMaxDepthLineEdit.text()
        if max_depth == 'None':
            max_depth = None
        else:
            try:
                max_depth = int(max_depth)
            except:
                self.error.emit('max_depth')
                return
            
            if max_depth <= 0 or float(self.app.rfMaxDepthLineEdit.text()) - max_depth != 0:
                self.error.emit('max_depth')
                return

        max_features = self.app.rfMaxFeaturesComboBox.currentText()
        if max_features == 'all':
            max_features = None

        class_weight = self.app.rfClassWeightComboBox.currentText()
        if class_weight == 'uniform':
            class_weight = None

        rf = ensemble.RandomForestClassifier(\
            n_estimators = self.app.rfNumEstimatorsSpinBox.value(), \
            criterion = self.app.rfCriterionComboBox.currentText(), \
            max_depth = max_depth, \
            min_samples_split = self.app.rfMinSamplesSplitSpinBox.value(), \
            min_samples_leaf = self.app.rfMinSamplesLeafSpinBox.value(), \
            max_features = max_features, \
            bootstrap = self.app.rfBootstrapCheckBox.isChecked(), \
            class_weight = class_weight, \
            random_state = 0)
        
        return RandomForest(rf, X, y, val_method=self.app.val_method, \
                val_size=self.app.holdoutSpinBox.value() / 100, \
                k = self.app.cvSpinBox.value(), \
                stratify=self.app.validationCheckBox.isChecked())

    def k_nearest_neighbors(self):
        """
        Trains and returns a k-nearest neighbors classifier.
        """
        data = self.app.data.train('integer')
        num_features = self.app.data.num_features('integer')
        X = data.iloc[:, 0: num_features]
        y = data.iloc[:, num_features]

        knn = neighbors.KNeighborsClassifier(\
            n_neighbors = self.app.knnNumNeighborsSpinBox.value(), \
            weights = self.app.knnWeightsComboBox.currentText(), \
            metric = self.app.knnMetricComboBox.currentText(), \
            algorithm = 'auto')
        
        return KNearestNeighbors(knn, X, y, val_method=self.app.val_method, \
                val_size=self.app.holdoutSpinBox.value() / 100, \
                k = self.app.cvSpinBox.value(), \
                stratify=self.app.validationCheckBox.isChecked())

    def logistic_regression(self):
        """
        Trains and returns a logistic regression classifier.
        """
        data = self.app.data.train('one-hot')
        num_features = self.app.data.num_features('one-hot')
        X = data.iloc[:, 0: num_features]
        y = data.iloc[:, num_features]

        try:
            penalty = float(self.app.lrRglrStrengthLineEdit.text())
        except:
            self.error.emit('penalty')
            return
        try:
            intercept_scaling = float(self.app.lrInterceptScalingLineEdit.text())
        except:
            self.error.emit('intercept_scaling')
            return
        try:
            tol = float(self.app.lrTolLineEdit.text())
        except:
            self.error.emit('tol')
            return
        try:
            max_iter = int(self.app.lrMaxIterLineEdit.text())
        except:
            self.error.emit('max_iter')
            return

        if penalty <= 0:
            self.error.emit('penalty')
            return
        if tol <= 0:
            self.error.emit('tol')
            return
        if max_iter <= 0 or float(self.app.lrMaxIterLineEdit.text()) - max_iter != 0:
            self.error.emit('max_iter')
            return

        class_weight = self.app.lrClassWeightComboBox.currentText()
        if class_weight == 'uniform':
            class_weight = None

        lr = linear_model.LogisticRegression(\
            penalty = self.app.lrRegularizationComboBox.currentText(), \
            tol = tol, \
            C = penalty, \
            fit_intercept = self.app.lrFitInterceptCheckBox.isChecked(), \
            intercept_scaling = intercept_scaling, \
            class_weight = class_weight, \
            solver = 'saga', \
            max_iter = max_iter, \
            multi_class = self.app.lrMultiClassComboBox.currentText(), \
            random_state = 0)
        
        try:
            model =  LogisticRegression(lr, X, y, val_method=self.app.val_method, \
                val_size=self.app.holdoutSpinBox.value() / 100, \
                k = self.app.cvSpinBox.value(), \
                stratify=self.app.validationCheckBox.isChecked())
        except exceptions.ConvergenceWarning:
            self.info.emit('converge')
            return

        return model

    def neural_network(self):
        """
        Trains and returns a neural network classifier.
        """
        data = self.app.data.train('one-hot')
        num_features = self.app.data.num_features('one-hot')
        X = data.iloc[:, 0: num_features]
        y = data.iloc[:, num_features]

        try:
            penalty = float(self.app.nnAlphaLineEdit.text())
        except:
            self.error.emit('penalty')
            return
        try:
            batch_size = int(self.app.nnBatchSizeLineEdit.text())
        except:
            self.error.emit('batch_size')
            return
        try:
            learning_rate_init = float(self.app.nnLearningRateInitLineEdit.text())
        except:
            self.error.emit('learning_rate_init')
            return
        try:
            tol = float(self.app.nnTolLineEdit.text())
        except:
            self.error.emit('tol')
            return
        try:
            max_iter = int(self.app.nnMaxIterLineEdit.text())
        except:
            self.error.emit('max_iter')
            return

        if penalty <= 0:
            self.error.emit('penalty')
            return
        if batch_size <= 0 or float(self.app.nnBatchSizeLineEdit.text()) - batch_size != 0:
            self.error.emit('batch_size')
            return
        if learning_rate_init <= 0:
            self.error.emit('learning_rate_init')
            return
        if tol <= 0:
            self.error.emit('tol')
            return
        if max_iter <= 0 or float(self.app.nnMaxIterLineEdit.text()) - max_iter != 0:
            self.error.emit('max_iter')
            return

        nn = neural_network.MLPClassifier(\
            hidden_layer_sizes = self.app.nnNumHiddenUnitsSpinBox.value(), \
            activation = self.app.nnActivationComboBox.currentText(), \
            solver = self.app.nnSolverComboBox.currentText(), \
            alpha = penalty, \
            batch_size = batch_size, \
            learning_rate = self.app.nnLearningRateComboBox.currentText(), \
            learning_rate_init = learning_rate_init, \
            max_iter = max_iter, \
            tol = tol, \
            early_stopping = self.app.nnEarlyStoppingCheckBox.isChecked(), \
            random_state = 0)

        try:
            model = NeuralNetwork(nn, X, y, val_method=self.app.val_method, \
                val_size=self.app.holdoutSpinBox.value() / 100, \
                k = self.app.cvSpinBox.value(), \
                stratify=self.app.validationCheckBox.isChecked())
        except exceptions.ConvergenceWarning:
            self.info.emit('converge')
            return

        return model

    def svm(self):
        """
        Trains and returns an SVM classifier.
        """
        data = self.app.data.train('one-hot')
        num_features = self.app.data.num_features('one-hot')
        X = data.iloc[:, 0: num_features]
        y = data.iloc[:, num_features]

        try:
            penalty = float(self.app.svmPenaltyLineEdit.text())
        except:
            self.error.emit('penalty')
            return
        try:
            gamma = float(self.app.svmGammaLineEdit.text())
        except:
            self.error.emit('kernel_coef')
            return
        try:
            coef0 = float(self.app.svmCoefLineEdit.text())
        except:
            self.error.emit('indenpendent_term')
            return
        try:
            tol = float(self.app.svmTolLineEdit.text())
        except:
            self.error.emit('tol')
            return
        try:
            max_iter = int(self.app.svmMaxIterLineEdit.text())
        except:
            self.error.emit('max_iter')
            return

        if penalty <= 0:
            self.error.emit('penalty')
            return
        if gamma <= 0:
            self.error.emit('kernel_coef')
            return
        if tol <= 0:
            self.error.emit('tol')
            return
        if max_iter <= 0 or float(self.app.svmMaxIterLineEdit.text()) - max_iter != 0:
            self.error.emit('max_iter')
            return

        class_weight = self.app.svmClassWeightComboBox.currentText()
        if class_weight == 'uniform':
            class_weight = None

        svc = svm.SVC(\
            C = penalty, \
            kernel = self.app.svmKernelComboBox.currentText(), \
            degree = self.app.svmDegreeSpinBox.value(), \
            gamma = gamma, \
            coef0 = coef0, \
            probability = True, \
            tol = tol, \
            class_weight = class_weight, \
            max_iter = max_iter, \
            random_state = 0)
           
        try: 
            model = SVM(svc, X, y, val_method=self.app.val_method, \
                val_size=self.app.holdoutSpinBox.value() / 100, \
                k = self.app.cvSpinBox.value(), \
                stratify=self.app.validationCheckBox.isChecked())
        except exceptions.ConvergenceWarning:
            self.info.emit('converge')
            return

        return model

    def naive_bayes(self):
        """
        Trains and returns a naive bayes classifier.
        """
        data = self.app.data.train('integer')
        num_features = self.app.data.num_features('integer')
        X = data.iloc[:, 0: num_features]
        y = data.iloc[:, num_features]

        class_prior = self.app.nbClassPriorLineEdit.text()
        if class_prior == 'None':
            class_prior = None
        else:
            try:
                class_prior = [float(i.strip()) for i in class_prior.split(',')]
            except:
                self.error.emit('class_prior')
                return

            if len(class_prior) != self.app.data.num_classes() or sum(class_prior) != 1:
                self.error.emit('class_prior')
                return

        if self.app.nbDistributionLabel.text() == 'multinomial':
            nb = naive_bayes.MultinomialNB(\
                alpha = self.app.nbAddSmoothDoubleSpinBox.value(), \
                fit_prior = self.app.nbFitPriorCheckBox.isChecked(), \
                class_prior = class_prior)
            mode = 'multinomial'
        elif self.app.nbDistributionLabel.text() == 'gaussian':
            nb = naive_bayes.GaussianNB(priors = class_prior)
            mode = 'gaussian'
        
        return NaiveBayes(nb, X, y, val_method=self.app.val_method, \
                val_size=self.app.holdoutSpinBox.value() / 100, \
                k = self.app.cvSpinBox.value(), \
                stratify=self.app.validationCheckBox.isChecked(), \
                mode=mode)

class App(QMainWindow):
    """
    Main window of the software.
    It inherits **QMainWindow**.
    """
    def __init__(self):
        """
        Initializes the GUI.
        """
        super().__init__()
        self.leftPanel = QStackedWidget(self)
        self.rightPanel = QGroupBox(self)
        self.initUI()

    def page_title(self, str, parent):
        """
        Returns a page title such as **Step 1: Select data**.
        Such a title appears at the top of each page in the left panel.
        It shows at which step the user is in the training process.

        :param str: a page title
        :type str: str
        :param parent: parent widget (current page in the left panel)
        :type parent: QWidget

        :returns: a page title (QLabel)
        """
        label = QLabel(str, parent)
        font = QFont()
        font.setPointSize(13)
        font.setBold(True)
        label.setFont(font)
        label.setAlignment(Qt.AlignCenter)
        return label

    def title(self, str, parent):
        """
        Returns bolded titles in each page of the left panel.
        Such a title appears at the top of each section in a page.
        It highlights what the user is intended to do.

        :param str: a section title
        :type str: str
        :param parent: parent widget (current page in the left panel)
        :type parent: QWidget

        :returns: a section title (QLabel)
        """
        label = QLabel(str, parent)
        font = QFont()
        font.setBold(True)
        label.setFont(font)
        return label

    def spin_box(self, val, min, max, stepsize, parent):
        """
        Return a spin box. 
        This spin box only allows integer values.

        :param val: initial value
        :type val: int
        :param min: minimum possible value
        :type min: int
        :param max: maximum possible value
        :type max: int
        :param stepsize: gap between neighboring values
        :type stepsize: int
        :param parent: parent widget
        :type parent: QWidget

        :returns: a spin box (QSpinBox)
        """
        box = QSpinBox(parent)
        box.setMinimum(min)
        box.setMaximum(max)
        box.setSingleStep(stepsize)
        box.setValue(val)
        return box

    # return a double spin box
    def double_spin_box(self, val, min, max, stepsize, prec, parent):
        """
        Return a double spin box.
        This spin box allows double values. 

        :param val: initial value
        :type val: float
        :param min: minimum possible value
        :type min: float
        :param max: maximum possible value
        :type max: float
        :param stepsize: gap between neighboring values
        :type stepsize: float
        :param prec: precision (number of decimal points)
        :type prec: int
        :param parent: parent widget
        :type parent: QWidget

        :returns: a spin box (QDoubleSpinBox)
        """
        box = QDoubleSpinBox(parent)
        box.setMinimum(min)
        box.setMaximum(max)
        box.setSingleStep(stepsize)
        box.setValue(val)
        box.setDecimals(prec)
        return box

    def load(self, labeled):
        """
        Load a data file. Process the data and show a summary of it.
        It checks the data format and raises an exception if the data is 
        in wrong format. Only .csv file with a header that contains 
        feature names are accepted. 

        If the data is labeled, the label column must be the last column, 
        and there must be at least 20 samples. An exception will be raised 
        if fewer than 20 samples are present. 

        If the data is unlabeled, the feature names must match those of the 
        labeled data. Extra white spaces are not allowed. An exception will
        be raised if the features do not match.

        :param labeled: indicates whether the data is labeled or not
        :type labeled: bool
        """
        path = QFileDialog.getOpenFileName(self.dataPage)[0]
        if path != '':
            # add labeled data.
            if labeled:
                try:
                    self.data = Data(path)
                # exception: wrong data format
                except:
                    self.error('format')
                    return
                # exception: too few samples (cutoff: 20)
                if self.data.num_samples() < 20:
                    self.error('num_samples')
                    return

                # once labeled data is successfully imported, 
                # enable subsequent operations.
                self.labeledFileDisplay.setText(self.data.name('labeled'))
                self.unlabeledFileDisplay.setText('')
                self.unlabeledFilePushButton.setEnabled(True)   # allow user to upload unlabeled data.
                self.splitFrame.setEnabled(True)                # allow user to split labeled data into training and test sets.
                self.validationFrame.setEnabled(True)           # allow user to select validation method.
                self.dataNextPushButton.setEnabled(True)        # allow user to proceed to next step (i.e. train classifiers).
                self.predictionPushButton.setEnabled(False)     # previously loaded unlabeled data is cleared

            # add unlabeled data (optional).
            else:
                try:
                    self.data.add_unlabeled_data(path)
                # exception: wrong data format
                except errors.ParserError:
                    self.error('format')
                    return
                # exception: features do not match
                except:
                    self.error('features')
                    return

                self.unlabeledFileDisplay.setText(self.data.name('unlabeled'))
                self.predictionPushButton.setEnabled(True)      # allow user to make predictions on the unlabeled data.

            self.data_summary()

    def data_summary(self):
        """
        Displays a summary of the loaded data.
        """
        self.data.summary(self.data_summary_)

    def set_decision_tree(self):
        """
        Sets hyperparameters of decision tree to default.
        """
        self.dtCriterionComboBox.setCurrentIndex(0)
        self.dtMaxDepthLineEdit.setText('None')
        self.dtMinSamplesSplitSpinBox.setValue(2)
        self.dtMinSamplesLeafSpinBox.setValue(1)
        self.dtClassWeightComboBox.setCurrentIndex(0)

    def set_random_forest(self):
        """
        Sets hyperparameter of random forest to default.
        """
        self.rfCriterionComboBox.setCurrentIndex(0)
        self.rfNumEstimatorsSpinBox.setValue(10)
        self.rfMaxFeaturesComboBox.setCurrentIndex(0)
        self.rfMaxDepthLineEdit.setText('None')
        self.rfMinSamplesSplitSpinBox.setValue(2)
        self.rfMinSamplesLeafSpinBox.setValue(1)
        self.rfBootstrapCheckBox.setChecked(True)
        self.rfClassWeightComboBox.setCurrentIndex(0)

    def set_k_nearest_neighbors(self):
        """
        Sets hyperparameters of k-nearest neighbors to default.
        The types of distances available depend on feature types.
        If the data has a mixture of discrete and continuous features,
        the user cannot train a k-nearest neighbors classifier.
        """
        self.knnNumNeighborsSpinBox.setValue(5)
        self.knnWeightsComboBox.setCurrentIndex(0)

        # discrete features: use hamming distance
        if self.data.feature_type() == 'discrete':
            self.classTypeListView.setRowHidden(3, False)
            self.knnMetricComboBox.setCurrentIndex(2)
            self.knnMetricListView.setRowHidden(0, True)
            self.knnMetricListView.setRowHidden(1, True)
            self.knnMetricListView.setRowHidden(2, False)
        # continuous features: use euclidean or manhatten distance
        elif self.data.feature_type() == 'continuous':
            self.classTypeListView.setRowHidden(3, False)
            self.knnMetricComboBox.setCurrentIndex(0)
            self.knnMetricListView.setRowHidden(0, False)
            self.knnMetricListView.setRowHidden(1, False)
            self.knnMetricListView.setRowHidden(2, True)
        # mixed features: DO NOT use KNN
        else:
            self.classTypeListView.setRowHidden(3, True)

    def set_logistic_regression(self):
        """
        Sets hyperparameters of logistic regression to default.
        """
        self.lrRegularizationComboBox.setCurrentIndex(0)
        self.lrRglrStrengthLineEdit.setText('1.0')
        self.lrFitInterceptCheckBox.setChecked(True)
        self.lrInterceptScalingLineEdit.setText('1.0')
        self.lrSolverComboBox.setCurrentIndex(0)
        self.lrMultiClassComboBox.setCurrentIndex(0)
        self.lrClassWeightComboBox.setCurrentIndex(0)
        self.lrTolLineEdit.setText('1e-3')
        self.lrMaxIterLineEdit.setText('500')

    def update_logistic_regression(self):
        """
        Sets hyperparameters of logistic regression according to 
        the selected solver. Some options are enabled/disabled.
        """
        if self.lrSolverComboBox.currentIndex() in [1, 2, 4]:
            self.lrRegularizationListView.setRowHidden(1, True)
            self.lrRegularizationComboBox.setCurrentIndex(0)
        else:
            self.lrRegularizationListView.setRowHidden(1, False)
        if self.lrSolverComboBox.currentIndex() == 0:
            self.lrMultiClassListView.setRowHidden(1, True)
            self.lrMultiClassComboBox.setCurrentIndex(0)
        else:
            self.lrMultiClassListView.setRowHidden(1, False)

    def set_neural_network(self):
        """
        Sets hyperparameters of neural network to default.
        """
        self.nnNumHiddenUnitsSpinBox.setValue(100)
        self.nnActivationComboBox.setCurrentIndex(0)
        self.nnSolverComboBox.setCurrentIndex(0)
        self.nnAlphaLineEdit.setText('1e-4')
        self.nnBatchSizeLineEdit.setText('20')
        self.nnLearningRateComboBox.setCurrentIndex(0)
        self.nnLearningRateInitLineEdit.setText('0.001')
        self.nnEarlyStoppingCheckBox.setChecked(False)
        self.nnTolLineEdit.setText('1e-3')
        self.nnMaxIterLineEdit.setText('500')

    def update_neural_network(self):
        """
        Set hyperparameters of neural network according to 
        the selected solver. Some options are enabled/disabled.
        """
        if self.nnSolverComboBox.currentIndex() == 2:
            self.nnLearningRateComboBox.setEnabled(True)
        else:
            self.nnLearningRateComboBox.setDisabled(True)
        if self.nnSolverComboBox.currentIndex() == 1:
            self.nnLearningRateInitLineEdit.setDisabled(True)
            self.nnEarlyStoppingCheckBox.setDisabled(True)
        else:
            self.nnLearningRateInitLineEdit.setEnabled(True)
            self.nnEarlyStoppingCheckBox.setEnabled(True)

    def set_svm(self):
        """
        Sets hyperparameters of SVM to default.
        """
        self.svmPenaltyLineEdit.setText('1.0')
        self.svmKernelComboBox.setCurrentIndex(0)
        self.svmDegreeSpinBox.setValue(3)
        gamma = max(0.001, 1 / self.data.num_features('one-hot'))
        self.svmGammaLineEdit.setText(str(round(gamma, 3)))
        self.svmCoefLineEdit.setText('0.0')
        self.svmClassWeightComboBox.setCurrentIndex(0)
        self.svmTolLineEdit.setText('1e-3')
        self.svmMaxIterLineEdit.setText('200')

    def update_svm(self):
        """
        Set hyperparameters of neural network according to 
        the selected kernel. Some options are enabled/disabled.
        """
        if self.svmKernelComboBox.currentIndex() == 2:
            self.svmDegreeSpinBox.setEnabled(True)
        else:
            self.svmDegreeSpinBox.setDisabled(True)
        if self.svmKernelComboBox.currentIndex() == 1:
            self.svmGammaLineEdit.setDisabled(True)
        else:
            self.svmGammaLineEdit.setEnabled(True)
        if self.svmKernelComboBox.currentIndex() in [0, 1]:
            self.svmCoefLineEdit.setDisabled(True)
        else:
            self.svmCoefLineEdit.setEnabled(True)

    def set_naive_bayes(self):
        """
        Sets hyperparameters of naive bayes to default.
        The type of distribution used depends on feature types.
        If the data has a mixture of discrete and continuous features,
        the user cannot train a naive bayes classifier.
        """
        # discrete features: use multinomial NB
        if self.data.feature_type() == 'discrete':
            self.classTypeListView.setRowHidden(7, False)
            self.nbDistributionLabel.setText('multinomial')
            self.nbAddSmoothDoubleSpinBox.setDisabled(False)
            self.nbFitPriorCheckBox.setDisabled(False)
            self.nbDoc.setText("<a href=\"http://scikit-learn.org/stable/modules/generated/sklearn.naive_bayes.MultinomialNB.html\">Documentation</a>")
        # continuous features: use gaussian NB
        elif self.data.feature_type() == 'continuous':
            self.classTypeListView.setRowHidden(7, False)
            self.nbDistributionLabel.setText('gaussian')
            self.nbAddSmoothDoubleSpinBox.setDisabled(True)
            self.nbFitPriorCheckBox.setDisabled(True)
            self.nbDoc.setText("<a href=\"http://scikit-learn.org/stable/modules/generated/sklearn.naive_bayes.GaussianNB.html\">Documentation</a>")
        # mixed features: DO NOT use NB
        else:
            self.classTypeListView.setRowHidden(7, True)

        self.nbAddSmoothDoubleSpinBox.setValue(1.0)
        self.nbFitPriorCheckBox.setChecked(True)
        self.nbClassPriorLineEdit.setText('None')

    def reset(self):
        """
        Resets hyperparameters of currently selected classifier type to default.
        """
        index = self.paramStack.currentIndex()
        if index == 1:      self.set_decision_tree()
        elif index == 2:    self.set_random_forest()
        elif index == 3:    self.set_k_nearest_neighbors()
        elif index == 4:    self.set_logistic_regression()
        elif index == 5:    self.set_neural_network()
        elif index == 6:    self.set_svm()
        elif index == 7:    self.set_naive_bayes()

        self.classNameLineEdit.clear()
        self.classCommentTextEdit.clear()

    def set(self):
        """
        Prepares data for training. This includes the following steps:

            - Encode data (integer and one-hot encoding).
            - Generate train/test split.
            - Decide validation method (holdout, cv or loo). 
            - Set classifier hyperparameters according to data.

        After this function is called, the software proceeds to the 2nd step 
        in training (i.e. train classifiers).
        """
        self.data.encode()  # encode data.

        # generate train/test split.
        test_size = self.splitSpinBox.value() / 100
        stratify = self.splitCheckBox.isChecked()
        self.data.split(test_size, stratify)

        # when data is 2D and continuous, allow plotting of data with decision regions.
        if self.data.feature_type() == 'continuous' \
            and self.data.num_features() == 2 and self.data.num_classes() in [2, 3]:
            self.dataPlotRadioButton.setEnabled(True)

        # decide validation method.
        val = 0
        if self.holdoutRadioButton.isChecked(): 
            self.val_method = 'holdout'
            if self.holdoutSpinBox.value() == 0:
                val = self.warn('holdout')  # warning: no validation data
        elif self.cvRadioButton.isChecked():    
            self.val_method = 'cv'
            if self.data.num_samples() / self.cvSpinBox.value() < 10:
                val = self.warn('cv')       # warning: too few samples in each fold
        elif self.looRadioButton.isChecked():   
            self.val_method = 'loo'
            if self.data.num_samples() > 50:
                val = self.warn('loo')      # warning: too many samples for leave-one-out

        if val == QMessageBox.Close:
            return

        # set classifier hyperparameters according to data.
        self.classTypeComboBox.setCurrentIndex(0)
        self.set_decision_tree()
        self.set_random_forest()
        self.set_k_nearest_neighbors()
        self.set_logistic_regression()
        self.set_neural_network()
        self.set_svm()
        self.set_naive_bayes()
        self.classNameLineEdit.clear()
        self.classCommentTextEdit.clear()
        self.trainStatusLabel.setText('')

        # if no data is held out for validation, only show metrics on training data.
        # this is for illustrating the importance of validation and should not be 
        # activated under other circumstances.
        if self.holdoutRadioButton.isChecked() and self.holdoutSpinBox.value() == 0:
            self.performanceComboBox.setCurrentIndex(1)
            self.performanceListView.setRowHidden(0, True)

        self.leftPanel.setCurrentIndex(1)   # proceed to the 2nd page.

    def clear(self, option):
        """
        Clears up changes made by user.

            - If option='all', the system is back to initial state.
              This is called by clicking on the finish button on the 3rd page.

            - If option='train', trained classifiers will be cleared.
              This is called by clicking on the return button on the 2nd page.
        
            - If option='test', the classifier selected for testing is no longer selected.
              This is called by clicking on the return button on the 3rd page.

        :params option: 'all', 'train' or 'test'
        :type params: str
        """

        # reset the software to initial state
        if option == 'all':
            self.clear('test')
            self.clear('models')
            self.data = None
            self.val_method = 'cv'
            self.tested = False
            self.unlabeledFilePushButton.setDisabled(True)
            self.labeledFileDisplay.setText('')
            self.unlabeledFileDisplay.setText('')
            self.data_summary_.clear()
            self.splitFrame.setDisabled(True)
            self.splitSpinBox.setValue(20)
            self.splitCheckBox.setChecked(True)
            self.validationFrame.setDisabled(True)
            self.holdoutSpinBox.setValue(20)
            self.holdoutSpinBox.setDisabled(True)
            self.cvRadioButton.setChecked(True)
            self.cvSpinBox.setValue(5)
            self.validationCheckBox.setChecked(True)
            self.testBackPushButton.setEnabled(True)

        # remove trained classifiers
        if option == 'models':
            self.curr_model = None
            self.models.clear()
            self.models_table.setRowCount(0)
            self.model_summary_.clear()
            self.classNextPushButton.setDisabled(True)
            self.performanceListView.setRowHidden(0, False)
            self.performanceComboBox.setCurrentIndex(0)
            self.canvas.figure.clear()
            self.canvas.draw()
            self.dataPlotRadioButton.setDisabled(True)
            self.confusionMatrixRadioButton.setChecked(True)
            self.leftPanel.setCurrentIndex(0)
            Model.clear()

        # reach here when user wants to change data during training
        if option == 'train':
            msg = 'The trained classifiers will be cleared. '
            msg += 'Do you want to proceed?'
            self.test_box.setText(msg)
            val = self.test_box.exec_()
            if val == QMessageBox.Cancel:
                return

            self.clear('models')

        # deselect classifier for testing
        elif option == 'test':
            self.selected_model = None
            self.bestPerformRadioButton.setChecked(True)
            self.metricComboBox.setCurrentIndex(0)
            self.leftPanel.setCurrentIndex(1)

    def train(self):
        """
        Trains a classifier in a new thread.
        """
        self.thread = Training_thread(self)
        self.thread.started.connect(self.start_train)
        self.thread.finished.connect(self.finish_train)
        self.thread.error.connect(self.error)
        self.thread.info.connect(self.info)
        self.thread.start()

    def start_train(self):
        """
        Starts training by freezing the 2nd page so that user cannot 
        alter the hyperparameters when training is in progress.
        """
        self.trainStatusLabel.setText('Training in progress...')
        self.classResetPushButton.setDisabled(True)
        self.classTrainPushButton.setDisabled(True)
        self.classNameLineEdit.setDisabled(True)
        self.classCommentTextEdit.setDisabled(True)
        self.classBackPushButton.setDisabled(True)
        self.classTypeComboBox.setDisabled(True)
        self.paramStack.setDisabled(True)

    def finish_train(self, model):
        """
        Receives the trained classifier from the training thread 
        and pushs the new classifier into the model table in the 
        right panel.

        :param model: trained classifier
        :type model: sklearn classifier object
        """
        # training failed (e.g. invalid hyperparameters or model did not converge)
        if model is None:
            self.trainStatusLabel.setText('Training failed.')

        else:
            self.trainStatusLabel.setText('Training completed.')
            
            # set user-supplied classifier name.
            name = self.classNameLineEdit.text().strip()
            if name != '':
                if name in self.models:
                    self.error('name')
                    return
                model.set_name(name)

            # set user-supplied cooment on classifier.
            model.set_comment(self.classCommentTextEdit.toPlainText().strip())

            self.classNameLineEdit.clear()
            self.classCommentTextEdit.clear()

            self.models[model.name()] = model   # add to the collection of classifiers.

            # show performance metrics with respect to the currently chosen data type.
            if self.performanceComboBox.currentIndex() == 0:
                self.push(model, 'val')
            else:
                self.push(model, 'train')

        # activate the 2nd page so that user can train new classifiers.
        self.classResetPushButton.setEnabled(True)
        self.classTrainPushButton.setEnabled(True)
        self.classNameLineEdit.setEnabled(True)
        self.classCommentTextEdit.setEnabled(True)
        self.classBackPushButton.setEnabled(True)
        self.classNextPushButton.setEnabled(True)
        self.classTypeComboBox.setEnabled(True)
        self.paramStack.setEnabled(True)

    def push(self, model, option):
        """
        Enters a classifier into the model table. Its performance metrics
        are added to the entry. Depending on _option_, it populates the 
        entry with appropriate performance metrics.

        :param model: trained classifier
        :type model: sklearn classifier object
        :param option: 'train' or 'val'
        :type option: str
        """
        if option == 'train':
            metrics = model.metrics('train')
        else:
            metrics = model.metrics('val')
            if metrics is None:
                return

        row = self.models_table.rowCount()  # enter as the last row
        self.models_table.insertRow(row)

        name_item = QTableWidgetItem(model.name())
        type_item = QTableWidgetItem(model.type())
        accuracy_item = QTableWidgetItem()
        precision_item = QTableWidgetItem()
        recall_item = QTableWidgetItem()
        f1_item = QTableWidgetItem()
        auroc_item = QTableWidgetItem()
        auprc_item = QTableWidgetItem()

        name_item.setFlags(Qt.ItemIsSelectable | Qt.ItemIsEnabled)
        type_item.setFlags(Qt.ItemIsSelectable | Qt.ItemIsEnabled)
        accuracy_item.setFlags(Qt.ItemIsSelectable | Qt.ItemIsEnabled)
        precision_item.setFlags(Qt.ItemIsSelectable | Qt.ItemIsEnabled)
        recall_item.setFlags(Qt.ItemIsSelectable | Qt.ItemIsEnabled)
        f1_item.setFlags(Qt.ItemIsSelectable | Qt.ItemIsEnabled)
        auroc_item.setFlags(Qt.ItemIsSelectable | Qt.ItemIsEnabled)
        auprc_item.setFlags(Qt.ItemIsSelectable | Qt.ItemIsEnabled)

        self.models_table.setItem(row, 0, name_item)
        self.models_table.setItem(row, 1, type_item)
        self.models_table.setItem(row, 2, accuracy_item)
        self.models_table.setItem(row, 3, precision_item)
        self.models_table.setItem(row, 4, recall_item)
        self.models_table.setItem(row, 5, f1_item)
        self.models_table.setItem(row, 6, auroc_item)
        self.models_table.setItem(row, 7, auprc_item)

        self.set_metrics(row, metrics)
        self.switch_model(row, 0)    # mark the new classifier as selected

    def set_metrics(self, row, metrics):
        """
        Enters the metrics in an entry.

        :param row: index of the entry
        :type row: int
        :param metrics: the collection of metrics to be entered
        :type metrics: ModelMetrics
        """
        self.models_table.item(row, 2).setText(str(metrics.accuracy()))
        self.models_table.item(row, 3).setText(str(metrics.precision()))
        self.models_table.item(row, 4).setText(str(metrics.recall()))
        self.models_table.item(row, 5).setText(str(metrics.f1()))
        self.models_table.item(row, 6).setText(str(metrics.auroc()))
        self.models_table.item(row, 7).setText(str(metrics.auprc()))

    def switch_metrics(self, index):
        """
        Switchs among metrics on training, validation and test data.
        The metric values in all table entries change accordingly.

        :param index: which set of metrics to display
        :type index: int
        """
        if index == 0:      option = 'val'
        elif index == 1:    option = 'train'
        else:               option = 'test'

        for i in range(0, self.models_table.rowCount()):
            model = self.models[self.models_table.item(i, 0).text()]
            metrics = model.metrics(option)
            if metrics is not None:
                self.set_metrics(i, metrics)
                self.models_table.showRow(i)
            else:
                self.models_table.hideRow(i)

        if self.curr_model is not None:
            row = 0
            for i in range(0, self.models_table.rowCount()):
                if self.models_table.item(i, 0).text() == self.curr_model.name():
                    row = i

            # if classifier has not been tested, clear model summary and plots
            # (useful when user switches from train/validation metrics to test metrics)
            if self.models_table.isRowHidden(row):
                self.model_summary_.clear()
                self.canvas.figure.clear()
                self.canvas.draw()
            else:
                self.plot()

        if self.leftPanel.currentIndex() == 2 and self.performanceComboBox.currentIndex() != 2:
            self.select('best')
            self.select('user')

    def switch_model(self, row, col):
        """
        Switchs among different classifiers in the table.
        A classifier summary and the plots for the newly highlighted classifier
        are generated.

        :param row: index of the selected entry
        :type row: int
        :param col: needed for parameter passing but not used
        :type col: int
        """
        self.models_table.setCurrentCell(row, 0)
        self.curr_model = self.models[self.models_table.item(row, 0).text()]
        self.model_summary()
        self.plot()

        # respond to mouse click when the classifier to be tested is user-picked
        if self.leftPanel.currentIndex() == 2 and self.userPickRadioButton.isChecked():
            self.selected_model = self.curr_model
            self.userPickLabel.setText(self.selected_model.name())
            self.testPushButton.setEnabled(True)

    def model_summary(self):
        """
        Displays a summary of the currently highlighted classifier.
        """
        self.curr_model.summary(self.model_summary_)

    def plot(self):
        """
        Shows plots of the currently highlighted classifier.
        """
        if self.curr_model is not None:
            index = self.performanceComboBox.currentIndex()
            if  index == 0:
                metrics = self.curr_model.metrics('val')
                option = 'train'
            elif index == 1:
                metrics = self.curr_model.metrics('train')
                option = 'train'
            else:
                metrics = self.curr_model.metrics('test')
                option = 'test'

            if self.confusionMatrixRadioButton.isChecked():
                metrics.plot_confusion_matrix(self.canvas)
            elif self.rocRadioButton.isChecked():
                metrics.plot_ROC(self.canvas)
            elif self.prRadioButton.isChecked():
                metrics.plot_precision_recall(self.canvas)
            elif self.dataPlotRadioButton.isChecked():
                self.curr_model.plot_decision_regions(option, self.canvas)

    def sort(self, col):
        """
        Sorts classifiers in descending order with respect to the selected 
        metric.

        :param col: index of the selected metric
        :type col: int
        """
        self.models_table.sortItems(col, Qt.DescendingOrder)

    def select(self, option):
        """
        Selects a classifier for testing.

            - If option='best', select the best-performing classifier based 
              on the selected metric.

            - If option='user', select the classifier highlighted by user.

        :param option: 'best' or 'user'
        :type option: str
        """
        if option == 'best' and self.bestPerformRadioButton.isChecked():
            index = self.metricComboBox.currentIndex()
            if index == 0:
                self.selected_model = None
                self.userPickLabel.setText('None')
                self.testPushButton.setDisabled(True)
                return
            else:
                col = index + 1
                self.sort(col)
                self.switch_model(0, 0)
                self.selected_model = self.curr_model        
                self.userPickLabel.setText(self.selected_model.name())
                self.testPushButton.setEnabled(True)
        elif option == 'user' and self.userPickRadioButton.isChecked():
            self.selected_model = self.curr_model        
            self.userPickLabel.setText(self.selected_model.name())
            self.testPushButton.setEnabled(True)

    def test(self):
        """
        Tests the selected classifier. Computes metrics on test data and 
        show them in the classifier's entry.
        """
        # when at least one classifier has been tested
        if self.tested:
            val = self.warn('test')
            if val == QMessageBox.Cancel:
                return
        # when no classifier has been tested
        else:
            msg = 'You cannot train more classifiers once you begin testing. '
            msg += 'Do you want to proceed?'
            self.test_box.setText(msg)
            val = self.test_box.exec_()
            if val == QMessageBox.Cancel:
                return

        if self.selected_model.type() in ['decision tree', 'random forest', \
            'k-nearest neighbors', 'naive bayes']:
            data = self.data.test('integer')
            num_features = self.data.num_features('integer')
        else:
            data = self.data.test('one-hot')
            num_features = self.data.num_features('one-hot')

        X = data.iloc[:, 0: num_features]
        y = data.iloc[:, num_features]
        self.selected_model.test(X, y)
        self.tested = True

        row = 0
        for i in range(0, self.models_table.rowCount()):
            if self.models_table.item(i, 0).text() == self.selected_model.name():
                row = i

        self.switch_model(row, 0)       # swtich to the selected model
        self.performanceListView.setRowHidden(2, False)
        self.bestPerformRadioButton.setChecked(True)
        self.metricComboBox.setCurrentIndex(0)          # DO NOT switch order
        self.performanceComboBox.setCurrentIndex(2)     # (bug in GUI library) 

        self.testBackPushButton.setDisabled(True)
        self.testFinishPushButton.setEnabled(True)

    def predict(self):
        """
        Makes prediction on new data using the selected classifier.
        Saves the prediction to a new file.
        """
        if self.curr_model.type() in ['decision tree', 'random forest', \
            'k-nearest neighbors', 'naive bayes']:
            data = self.data.prediction('integer')
        else:
            data = self.data.prediction('one-hot')

        path = QFileDialog.getSaveFileName(self.testPage)[0]
        self.curr_model.predict(data, path)

    def finish(self):
        """
        Finishs analyzing the current datasets.
        Asks user for future action (quit or start a new analysis).
        """
        msg = 'Do you want to analyze more data?'
        self.finish_box.setText(msg)
        val = self.finish_box.exec_()
        if val == QMessageBox.Yes:
            self.clear('all')
        elif val == QMessageBox.No:
            QApplication.instance().quit()

    def info(self, flag):
        """
        Shows reminders in a message box.

        :param flag: indicates the event that triggers the reminder.
        :type flag: str
        """
        if flag == 'converge':
            msg = 'Maximum iterations reached and the classifier has not converged yet. '
            msg += 'Consider increasing max_iter.'
        self.info_box.setText(msg)
        self.info_box.exec_()

    def warn(self, flag):
        """
        Shows warning messages in a message box.

        :param flag: indicates the event that triggers the warning.
        :type flag: str
        """
        self.warn_message(flag)
        return self.warn_box.exec_()

    def error(self, flag):
        """
        Shows error messages in a message box.

        :param flag: indicates the event that triggers the error.
        :type flag: str
        """
        self.error_message(flag)
        return self.err_box.exec_()

    def warn_message(self, flag):
        """
        Constructs warning messages.

        :param flag: indicates the event that triggers the warning.
        :type flag: str
        """
        if flag == 'holdout':
            msg = 'No validation data is allocated. Classifiers may overfit.'
        elif flag == 'cv':
            msg = 'Too few samples in each fold. '
            msg += 'Use fewer folds or consider leave-one-out validation.'
        elif flag == 'loo':
            msg = 'Too many samples for leave-one-out validation. '
            msg += 'Consider hold-out validation or k-fold cross-validation.'
        elif flag == 'test':
            msg = 'Model selection on test data may lead to an overfit model.'
        self.warn_box.setText(msg)

    def error_message(self, flag):
        """
        Constructs error messages.

        :param flag: indicates the event that triggers the error.
        :type flag: str
        """
        if flag == 'format':
            msg = 'Wrong data format. Only .csv is accepted.'
        elif flag == 'num_samples':
            msg = 'Too few samples. At least 20 samples are required.'
        elif flag == 'features':
            msg = 'Feature names do not match.'
        elif flag == 'max_depth':
            msg = 'max_depth: must be a positive integer (e.g. 3) or None'
        elif flag == 'penalty':
            msg = 'C/alpha: must be a positive number (e.g. 0.5)'
        elif flag == 'intercept_scaling':
            msg = 'intercept_scaling: must be a number (e.g. 1.0)'
        elif flag == 'batch_size':
            msg = 'batch_size: must be a positive integer (e.g. 200)'
        elif flag == 'learning_rate_init':
            msg = 'learning_rate_init: must be a positive number (e.g. 0.001)'
        elif flag == 'kernel_coef':
            msg = 'gamma: must be a positive number (e.g. 0.5)'
        elif flag == 'indenpendent_term':
            msg = 'coef0: must be a number (e.g. 0)'
        elif flag == 'class_prior':
            msg = 'class_prior: must be None or follow the format <value>,...,<value> '
            msg += 'where the number of values equals the number of classes '
            msg += 'and the values sum to 1 (e.g. 0.2,0.6,0.2 for three classes)'
        elif flag == 'tol':
            msg = 'tol: must be a positive number (e.g. 1e-4)'
        elif flag == 'max_iter':
            msg = 'max_iter: must be a positive integer (e.g. 200)'
        elif flag == 'name':
            msg = 'Classifier name already exists.'
        self.err_box.setText(msg)

    def initUI(self):
        """
        Sets up the GUI.
        """
        self.data = None
        self.val_method = 'cv'
        self.models = dict()
        self.curr_model = None      # classifier highlighted in the table
        self.selected_model = None  # classifier selected for testing
        self.tested = False

        self.info_box = QMessageBox()
        self.info_box.setIcon(QMessageBox.Information)
        self.info_box.setStandardButtons(QMessageBox.Ok)

        self.warn_box = QMessageBox()
        self.warn_box.setIcon(QMessageBox.Warning)
        self.warn_box.setStandardButtons(QMessageBox.Close | QMessageBox.Ignore)
        
        self.err_box = QMessageBox()
        self.err_box.setIcon(QMessageBox.Critical)
        self.err_box.setStandardButtons(QMessageBox.Ok)

        self.test_box = QMessageBox()
        self.test_box.setIcon(QMessageBox.Question)
        self.test_box.setStandardButtons(QMessageBox.Cancel | QMessageBox.Yes)

        self.finish_box = QMessageBox()
        self.finish_box.setIcon(QMessageBox.Question)
        self.finish_box.setStandardButtons(QMessageBox.Cancel | QMessageBox.No | QMessageBox.Yes)

        self.font = QFont()         # font for table and summary
        self.font.setPointSize(10)

        ########## 1st page: data loading and splitting ##########
        self.dataPage = QWidget()
        self.openIcon = QIcon('ml4bio/icons/open.png')
        self.dataPageLayout = QVBoxLayout(self.dataPage)
        self.dataPageTitle = self.page_title('Step 1: Select Data', self.dataPage)
        self.dataPageLayout.addWidget(self.dataPageTitle)
        self.dataPageLine = QFrame(self.dataPage)
        self.dataPageLine.setFrameShape(QFrame.HLine)
        self.dataPageLine.setFrameShadow(QFrame.Sunken)
        self.dataPageLayout.addWidget(self.dataPageLine)

        ### load labeled data
        self.labeledDataLabel = self.title('Labeled Data (.csv):', self.dataPage)
        self.labeledFilePushButton = QPushButton('Select File...', self.dataPage)
        self.labeledFilePushButton.setIcon(self.openIcon)
        self.labeledFilePushButton.setMaximumWidth(140)
        self.labeledFileSpacer = QSpacerItem(40, 20)
        self.labeledFileDisplay = QLabel('', self.dataPage)

        self.dataPageLayout.addWidget(self.labeledDataLabel)
        self.labeledFileLayout = QHBoxLayout()
        self.labeledFileLayout.addWidget(self.labeledFilePushButton)
        self.labeledFileLayout.addItem(self.labeledFileSpacer)
        self.labeledFileLayout.addWidget(self.labeledFileDisplay)
        self.dataPageLayout.addLayout(self.labeledFileLayout)

        ### load unlabeled data
        self.unlabeledDataLabel = self.title('Unlabeled Data (.csv):', self.dataPage)
        self.unlabeledFilePushButton = QPushButton('Select File...', self.dataPage)
        self.unlabeledFilePushButton.setIcon(self.openIcon)
        self.unlabeledFilePushButton.setMaximumWidth(140)
        self.unlabeledFilePushButton.setDisabled(True)
        self.unlabeledFileSpacer = QSpacerItem(40, 20)
        self.unlabeledFileDisplay = QLabel('', self.dataPage)

        self.dataPageLayout.addWidget(self.unlabeledDataLabel)
        self.unlabeledFileLayout = QHBoxLayout()
        self.unlabeledFileLayout.addWidget(self.unlabeledFilePushButton)
        self.unlabeledFileLayout.addItem(self.unlabeledFileSpacer)
        self.unlabeledFileLayout.addWidget(self.unlabeledFileDisplay)
        self.dataPageLayout.addLayout(self.unlabeledFileLayout)

        ### data summary
        self.dataSummaryLabel = self.title('Data Summary:', self.dataPage)
        self.data_summary_ = QTreeWidget(self.dataPage)
        self.data_summary_.setColumnCount(2)
        self.data_summary_.setHeaderHidden(True)
        self.data_summary_.setColumnWidth(0, 200)
        self.dataPageLayout.addWidget(self.dataSummaryLabel)
        self.dataPageLayout.addWidget(self.data_summary_)

        ### train/test split
        self.trainTestLabel = self.title('Train/Test Split:', self.dataPage)
        self.splitFrame = QFrame(self.dataPage)
        self.splitFrame.setAutoFillBackground(True)
        self.splitFrame.setDisabled(True)
        self.splitSpinBox = self.spin_box(20, 10, 50, 10, self.splitFrame)
        self.splitSpinBox.setMaximumWidth(60)
        self.splitLabel = QLabel('% for test', self.splitFrame)
        self.splitCheckBox = QCheckBox('Stratified Sampling', self.splitFrame)
        self.splitCheckBox.setChecked(True)

        self.dataPageLayout.addWidget(self.trainTestLabel)
        self.trainTestLayout = QHBoxLayout()
        self.trainTestLayout.addWidget(self.splitSpinBox)
        self.trainTestLayout.addWidget(self.splitLabel)
        self.trainTestLayout.addWidget(self.splitCheckBox)
        self.splitLayout = QVBoxLayout(self.splitFrame)
        self.splitLayout.addLayout(self.trainTestLayout)
        self.dataPageLayout.addWidget(self.splitFrame)

        ### validation methodology
        self.validationLabel = self.title('Validation Methodology:', self.dataPage)
        self.validationFrame = QFrame(self.dataPage)
        self.validationFrame.setAutoFillBackground(True)
        self.validationFrame.setDisabled(True)
        self.holdoutRadioButton = QRadioButton('Holdout Validation', self.validationFrame)
        self.cvRadioButton = QRadioButton('K-Fold Cross-Validation', self.validationFrame)
        self.cvRadioButton.setChecked(True)
        self.looRadioButton = QRadioButton('Leave-One-Out Validation', self.validationFrame)
        self.holdoutSpacer = QSpacerItem(40, 20, QSizePolicy.Expanding, QSizePolicy.Minimum)
        self.cvSpacer = QSpacerItem(40, 20, QSizePolicy.Expanding, QSizePolicy.Minimum)
        self.looSpacer = QSpacerItem(40, 20, QSizePolicy.Expanding, QSizePolicy.Minimum)
        self.validationSpacer = QSpacerItem(40, 20, QSizePolicy.Expanding, QSizePolicy.Minimum)
        self.holdoutSpinBox = self.spin_box(20, 0, 50, 10, self.validationFrame)
        self.holdoutSpinBox.setMaximumWidth(50)
        self.holdoutSpinBox.setDisabled(True)
        self.cvSpinBox = self.spin_box(5, 2, 10, 1, self.validationFrame)
        self.validationCheckBox = QCheckBox('Stratified Sampling', self.validationFrame)
        self.validationCheckBox.setChecked(True)
        self.holdoutLabel = QLabel('% for validation', self.validationFrame)
        self.cvLabel = QLabel('folds', self.validationFrame)

        self.dataPageLayout.addWidget(self.validationLabel)
        self.holdoutLayout = QHBoxLayout()
        self.holdoutLayout.addWidget(self.holdoutRadioButton)
        self.holdoutLayout.addItem(self.holdoutSpacer)
        self.holdoutLayout.addWidget(self.holdoutSpinBox)
        self.holdoutLayout.addWidget(self.holdoutLabel)
        self.cvLayout = QHBoxLayout()
        self.cvLayout.addWidget(self.cvRadioButton)
        self.cvLayout.addItem(self.cvSpacer)
        self.cvLayout.addWidget(self.cvSpinBox)
        self.cvLayout.addWidget(self.cvLabel)
        self.looLayout = QHBoxLayout()
        self.looLayout.addWidget(self.looRadioButton)
        self.looLayout.addItem(self.looSpacer)
        self.samplingLayout = QHBoxLayout()
        self.samplingLayout.addItem(self.validationSpacer)
        self.samplingLayout.addWidget(self.validationCheckBox)
        self.validationLayout = QVBoxLayout(self.validationFrame)
        self.validationLayout.addLayout(self.holdoutLayout)
        self.validationLayout.addLayout(self.cvLayout)
        self.validationLayout.addLayout(self.looLayout)
        self.validationLayout.addLayout(self.samplingLayout)
        self.dataPageLayout.addWidget(self.validationFrame)

        ### data spacer and line
        self.dataSpacer = QSpacerItem(40, 20, QSizePolicy.Minimum, QSizePolicy.Expanding)
        self.dataPageLayout.addItem(self.dataSpacer)
        self.dataLine = QFrame(self.dataPage)
        self.dataLine.setFrameShape(QFrame.HLine)
        self.dataLine.setFrameShadow(QFrame.Sunken)
        self.dataPageLayout.addWidget(self.dataLine)

        ### next button
        self.dataNextSpacer = QSpacerItem(40, 20, QSizePolicy.Expanding, QSizePolicy.Minimum)
        self.dataNextPushButton = QPushButton('Next', self.dataPage)
        self.dataNextPushButton.setDefault(True)
        self.dataNextPushButton.setMinimumWidth(90)
        self.dataNextPushButton.setDisabled(True)
        self.dataNextLayout = QHBoxLayout()
        self.dataNextLayout.addItem(self.dataNextSpacer)
        self.dataNextLayout.addWidget(self.dataNextPushButton)
        self.dataPageLayout.addLayout(self.dataNextLayout)

        self.leftPanel.addWidget(self.dataPage)

        # signal handling
        self.labeledFilePushButton.clicked.connect(lambda: self.load(labeled=True))
        self.unlabeledFilePushButton.clicked.connect(lambda: self.load(labeled=False))

        self.holdoutRadioButton.toggled.connect(self.holdoutSpinBox.setEnabled)
        self.holdoutRadioButton.toggled.connect(self.cvSpinBox.setDisabled)
        self.holdoutRadioButton.toggled.connect(self.validationCheckBox.setEnabled)
        self.cvRadioButton.toggled.connect(self.holdoutSpinBox.setDisabled)
        self.cvRadioButton.toggled.connect(self.cvSpinBox.setEnabled)
        self.cvRadioButton.toggled.connect(self.validationCheckBox.setEnabled)
        self.looRadioButton.toggled.connect(self.holdoutSpinBox.setDisabled)
        self.looRadioButton.toggled.connect(self.cvSpinBox.setDisabled)
        self.looRadioButton.toggled.connect(self.validationCheckBox.setDisabled)

        self.dataNextPushButton.clicked.connect(self.set)
        
        ########## 2nd page: training ##########

        self.modelPage = QWidget()
        self.modelPageLayout = QVBoxLayout(self.modelPage)
        self.modelPageTitle = self.page_title('Step 2: Train Classifiers', self.modelPage)
        self.modelPageLayout.addWidget(self.modelPageTitle)
        self.modelPageLine = QFrame(self.modelPage)
        self.modelPageLine.setFrameShape(QFrame.HLine)
        self.modelPageLine.setFrameShadow(QFrame.Sunken)
        self.modelPageLayout.addWidget(self.modelPageLine)

        ### classifier type
        self.classTypeLabel = self.title('Classifier Type:', self.modelPage)
        self.classTypeComboBox = QComboBox(self.modelPage)
        self.classTypeListView = self.classTypeComboBox.view()
        self.classTypeLayout = QHBoxLayout()
        self.classTypeLayout.addWidget(self.classTypeLabel)
        self.classTypeLayout.addWidget(self.classTypeComboBox)

        ### classifier parameter stack
        self.paramStack = QStackedWidget(self.modelPage)
        self.paramStack.setMinimumHeight(320)
        self.modelPageLayout.addLayout(self.classTypeLayout)
        self.modelPageLayout.addWidget(self.paramStack)

        ## initial empty page
        self.noneIcon = QIcon('ml4bio/icons/none.png')
        self.classTypeComboBox.addItem(self.noneIcon, '-------- select --------')
        self.initPage = QWidget()
        self.paramStack.addWidget(self.initPage)

        ## decision tree
        self.dtIcon = QIcon('ml4bio/icons/dt.png')
        self.classTypeComboBox.addItem(self.dtIcon, 'Decision Tree')
        self.dtPage = QWidget()
        self.paramStack.addWidget(self.dtPage)

        # fields
        self.dtCriterionComboBox = QComboBox(self.dtPage)
        self.dtCriterionComboBox.addItem('gini')
        self.dtCriterionComboBox.addItem('entropy')
        self.dtMaxDepthLineEdit = QLineEdit('None', self.dtPage)
        self.dtMinSamplesSplitSpinBox = self.spin_box(2, 2, 20, 1, self.dtPage)
        self.dtMinSamplesLeafSpinBox = self.spin_box(1, 1, 20, 1, self.dtPage)
        self.dtClassWeightComboBox = QComboBox(self.dtPage)
        self.dtClassWeightComboBox.addItem('uniform')
        self.dtClassWeightComboBox.addItem('balanced')

        self.dtDoc = QLabel("<a href=\"http://scikit-learn.org/stable/modules/generated/sklearn.tree.DecisionTreeClassifier.html\">Documentation</a>")
        self.dtDoc.setAlignment(Qt.AlignRight)
        self.dtDoc.setOpenExternalLinks(True)

        # layout
        self.dtLayout = QFormLayout()
        self.dtLayout.addRow('criterion:', self.dtCriterionComboBox)
        self.dtLayout.addRow('max_depth:', self.dtMaxDepthLineEdit)
        self.dtLayout.addRow('min_samples_split:', self.dtMinSamplesSplitSpinBox)
        self.dtLayout.addRow('min_samples_leaf:', self.dtMinSamplesLeafSpinBox)
        self.dtLayout.addRow('class_weight:', self.dtClassWeightComboBox)

        self.dtPageLayout = QVBoxLayout(self.dtPage)
        self.dtPageLayout.addLayout(self.dtLayout)

        self.dtSpacer = QSpacerItem(40, 20, QSizePolicy.Minimum, QSizePolicy.Expanding)
        self.dtPageLayout.addItem(self.dtSpacer)
        self.dtPageLayout.addWidget(self.dtDoc)

        ## Random forest
        self.rfIcon = QIcon('ml4bio/icons/rf.png')
        self.classTypeComboBox.addItem(self.rfIcon, 'Random Forest')
        self.rfPage = QWidget()
        self.paramStack.addWidget(self.rfPage)

        # fields
        self.rfCriterionComboBox = QComboBox(self.rfPage)
        self.rfCriterionComboBox.addItem('gini')
        self.rfCriterionComboBox.addItem('entropy')
        self.rfNumEstimatorsSpinBox = self.spin_box(10, 2, 20, 1, self.rfPage)
        self.rfMaxFeaturesComboBox = QComboBox(self.rfPage)
        self.rfMaxFeaturesComboBox.addItem('sqrt')
        self.rfMaxFeaturesComboBox.addItem('log2')
        self.rfMaxFeaturesComboBox.addItem('all')
        self.rfMaxDepthLineEdit = QLineEdit('None', self.rfPage)
        self.rfMinSamplesSplitSpinBox = self.spin_box(2, 2, 20, 1, self.rfPage)
        self.rfMinSamplesLeafSpinBox = self.spin_box(1, 1, 20, 1, self.rfPage)
        self.rfBootstrapCheckBox = QCheckBox('', self.rfPage)
        self.rfBootstrapCheckBox.setChecked(True)
        self.rfClassWeightComboBox = QComboBox(self.rfPage)
        self.rfClassWeightComboBox.addItem('uniform')
        self.rfClassWeightComboBox.addItem('balanced')

        self.rfDoc = QLabel("<a href=\"http://scikit-learn.org/stable/modules/generated/sklearn.ensemble.RandomForestClassifier.html\">Documentation</a>")
        self.rfDoc.setAlignment(Qt.AlignRight)
        self.rfDoc.setOpenExternalLinks(True)

        # layout
        self.rfLayout = QFormLayout()
        self.rfLayout.addRow('criterion:', self.rfCriterionComboBox)
        self.rfLayout.addRow('n_estimators:', self.rfNumEstimatorsSpinBox)
        self.rfLayout.addRow('max_features:', self.rfMaxFeaturesComboBox)
        self.rfLayout.addRow('max_depth:', self.rfMaxDepthLineEdit)
        self.rfLayout.addRow('min_samples_split:', self.rfMinSamplesSplitSpinBox)
        self.rfLayout.addRow('min_samples_leaf:', self.rfMinSamplesLeafSpinBox)
        self.rfLayout.addRow('bootstrap:', self.rfBootstrapCheckBox)
        self.rfLayout.addRow('class_weight:', self.rfClassWeightComboBox)

        self.rfPageLayout = QVBoxLayout(self.rfPage)
        self.rfPageLayout.addLayout(self.rfLayout)

        self.rfSpacer = QSpacerItem(40, 20, QSizePolicy.Minimum, QSizePolicy.Expanding)
        self.rfPageLayout.addItem(self.rfSpacer)
        self.rfPageLayout.addWidget(self.rfDoc)

        ## K-nearest neighbors
        self.knnIcon = QIcon('ml4bio/icons/knn.png')
        self.classTypeComboBox.addItem(self.knnIcon, 'K-Nearest Neighbors')
        self.knnPage = QWidget()
        self.paramStack.addWidget(self.knnPage)

        # fields
        self.knnNumNeighborsSpinBox = self.spin_box(5, 1, 20, 1, self.knnPage)
        self.knnWeightsComboBox = QComboBox(self.knnPage)
        self.knnWeightsComboBox.addItem('uniform')
        self.knnWeightsComboBox.addItem('distance')
        self.knnMetricComboBox = QComboBox(self.knnPage)
        self.knnMetricListView =self.knnMetricComboBox.view()
        self.knnMetricComboBox.addItem('euclidean') #
        self.knnMetricComboBox.addItem('manhattan') # DO NOT change order
        self.knnMetricComboBox.addItem('hamming')   #

        self.knnDoc = QLabel("<a href=\"http://scikit-learn.org/stable/modules/generated/sklearn.neighbors.KNeighborsClassifier.html\">Documentation</a>")
        self.knnDoc.setAlignment(Qt.AlignRight)
        self.knnDoc.setOpenExternalLinks(True)

        # layout
        self.knnLayout = QFormLayout()
        self.knnLayout.addRow('n_neighbors:', self.knnNumNeighborsSpinBox)
        self.knnLayout.addRow('weights:', self.knnWeightsComboBox)
        self.knnLayout.addRow('metric:', self.knnMetricComboBox)

        self.knnPageLayout = QVBoxLayout(self.knnPage)
        self.knnPageLayout.addLayout(self.knnLayout)

        self.knnSpacer = QSpacerItem(40, 20, QSizePolicy.Minimum, QSizePolicy.Expanding)
        self.knnPageLayout.addItem(self.knnSpacer)
        self.knnPageLayout.addWidget(self.knnDoc)

        ## Logistic regression
        self.lrIcon = QIcon('ml4bio/icons/lr.png')
        self.classTypeComboBox.addItem(self.lrIcon, 'Logistic Regression')
        self.lrPage = QWidget()
        self.paramStack.addWidget(self.lrPage)

        # fields
        self.lrRegularizationComboBox = QComboBox(self.lrPage)
        self.lrRegularizationComboBox.addItem('l2')
        self.lrRegularizationComboBox.addItem('l1')
        self.lrRegularizationListView = self.lrRegularizationComboBox.view()
        self.lrRglrStrengthLineEdit = QLineEdit('1.0', self.lrPage)
        self.lrFitInterceptCheckBox = QCheckBox('', self.lrPage)
        self.lrFitInterceptCheckBox.setChecked(True)
        self.lrInterceptScalingLineEdit = QLineEdit('1.0', self.lrPage)
        self.lrSolverComboBox = QComboBox(self.lrPage)
        self.lrSolverComboBox.addItem('liblinear')
        self.lrSolverComboBox.addItem('lbfgs')
        self.lrSolverComboBox.addItem('sag')
        self.lrSolverComboBox.addItem('saga')
        self.lrSolverComboBox.addItem('newton-cg')
        self.lrMultiClassComboBox = QComboBox(self.lrPage)
        self.lrMultiClassComboBox.addItem('ovr')
        self.lrMultiClassComboBox.addItem('multinomial')
        self.lrMultiClassListView = self.lrMultiClassComboBox.view()
        self.lrClassWeightComboBox = QComboBox(self.lrPage)
        self.lrClassWeightComboBox.addItem('uniform')
        self.lrClassWeightComboBox.addItem('balanced')

        self.lrStopLabel = QLabel('Stopping Criteria:', self.lrPage)
        self.lrTolLabel = QLabel('tol:', self.lrPage)
        self.lrTolLabel.setMinimumWidth(60)
        self.lrTolLineEdit = QLineEdit('1e-3', self.lrPage)
        self.lrMaxIterLabel = QLabel('max_iter:', self.lrPage)
        self.lrMaxIterLabel.setMinimumWidth(60)
        self.lrMaxIterLineEdit = QLineEdit('500', self.lrPage)

        self.lrDoc = QLabel("<a href=\"http://scikit-learn.org/stable/modules/generated/sklearn.linear_model.LogisticRegression.html\">Documentation</a>")
        self.lrDoc.setAlignment(Qt.AlignRight)
        self.lrDoc.setOpenExternalLinks(True)

        # layout
        self.lrLayout = QFormLayout()
        self.lrLayout.addRow('penalty:', self.lrRegularizationComboBox)
        self.lrLayout.addRow('C:', self.lrRglrStrengthLineEdit)
        self.lrLayout.addRow('fit_intercept:', self.lrFitInterceptCheckBox)
        self.lrLayout.addRow('intercept_scaling:', self.lrInterceptScalingLineEdit)
        self.lrLayout.addRow('solver:', self.lrSolverComboBox)
        self.lrLayout.addRow('multi_class:', self.lrMultiClassComboBox)
        self.lrLayout.addRow('class_weight:', self.lrClassWeightComboBox)

        self.lrSpacer = QSpacerItem(40, 20, QSizePolicy.Minimum, QSizePolicy.Expanding)

        self.lrStopLayout = QHBoxLayout()
        self.lrStopLayout.addWidget(self.lrTolLabel)
        self.lrStopLayout.addWidget(self.lrTolLineEdit)
        self.lrStopLayout.addWidget(self.lrMaxIterLabel)
        self.lrStopLayout.addWidget(self.lrMaxIterLineEdit)

        self.lrPageLayout = QVBoxLayout(self.lrPage)
        self.lrPageLayout.addLayout(self.lrLayout)
        self.lrPageLayout.addItem(self.lrSpacer)
        self.lrPageLayout.addWidget(self.lrStopLabel)
        self.lrPageLayout.addLayout(self.lrStopLayout)

        self.lrPageLayout.addWidget(self.lrDoc)

        ## Neural Network
        self.nnIcon = QIcon('ml4bio/icons/nn.png')
        self.classTypeComboBox.addItem(self.nnIcon, 'Neural Network')
        self.nnPage = QWidget()
        self.paramStack.addWidget(self.nnPage)

        # fields
        self.nnNumHiddenUnitsSpinBox = self.spin_box(100, 1, 200, 1, self.nnPage)
        self.nnActivationComboBox = QComboBox(self.nnPage)
        self.nnActivationComboBox.addItem('relu')
        self.nnActivationComboBox.addItem('logistic')
        self.nnActivationComboBox.addItem('tanh')
        self.nnActivationComboBox.addItem('identity')
        self.nnSolverComboBox = QComboBox(self.nnPage)
        self.nnSolverComboBox.addItem('adam')
        self.nnSolverComboBox.addItem('lbfgs')
        self.nnSolverComboBox.addItem('sgd')
        self.nnAlphaLineEdit = QLineEdit('1e-4', self.nnPage)
        self.nnBatchSizeLineEdit = QLineEdit('20', self.nnPage)
        self.nnLearningRateComboBox = QComboBox(self.nnPage)
        self.nnLearningRateComboBox.addItem('constant')
        self.nnLearningRateComboBox.addItem('invscaling')
        self.nnLearningRateComboBox.addItem('adaptive')
        self.nnLearningRateInitLineEdit = QLineEdit('0.001', self.nnPage)
        self.nnEarlyStoppingCheckBox = QCheckBox('', self.nnPage)

        self.nnStopLabel = QLabel('Stopping Criteria:', self.nnPage)
        self.nnTolLabel = QLabel('tol:', self.nnPage)
        self.nnTolLabel.setMinimumWidth(60)
        self.nnTolLineEdit = QLineEdit('1e-3', self.nnPage)
        self.nnMaxIterLabel = QLabel('max_iter:', self.nnPage)
        self.nnMaxIterLabel.setMinimumWidth(60)
        self.nnMaxIterLineEdit = QLineEdit('500', self.nnPage)

        self.nnDoc = QLabel("<a href=\"http://scikit-learn.org/stable/modules/generated/sklearn.neural_network.MLPClassifier.html\">Documentation</a>")
        self.nnDoc.setAlignment(Qt.AlignRight)
        self.nnDoc.setOpenExternalLinks(True)

        # layout
        self.nnLayout = QFormLayout()
        self.nnLayout.addRow('num_hidden_units:', self.nnNumHiddenUnitsSpinBox)
        self.nnLayout.addRow('activation:', self.nnActivationComboBox)
        self.nnLayout.addRow('solver:', self.nnSolverComboBox)
        self.nnLayout.addRow('alpha:', self.nnAlphaLineEdit)
        self.nnLayout.addRow('batch_size:', self.nnBatchSizeLineEdit)
        self.nnLayout.addRow('learning_rate:', self.nnLearningRateComboBox)
        self.nnLayout.addRow('learning_rate_init:', self.nnLearningRateInitLineEdit)
        self.nnLayout.addRow('early_stopping:', self.nnEarlyStoppingCheckBox)

        self.nnSpacer = QSpacerItem(40, 20, QSizePolicy.Minimum, QSizePolicy.Expanding)

        self.nnStopLayout = QHBoxLayout()
        self.nnStopLayout.addWidget(self.nnTolLabel)
        self.nnStopLayout.addWidget(self.nnTolLineEdit)
        self.nnStopLayout.addWidget(self.nnMaxIterLabel)
        self.nnStopLayout.addWidget(self.nnMaxIterLineEdit)

        self.nnPageLayout = QVBoxLayout(self.nnPage)
        self.nnPageLayout.addLayout(self.nnLayout)
        self.nnPageLayout.addItem(self.nnSpacer)
        self.nnPageLayout.addWidget(self.nnStopLabel)
        self.nnPageLayout.addLayout(self.nnStopLayout)

        self.nnPageLayout.addWidget(self.nnDoc)

        ## SVM
        self.svmIcon = QIcon('ml4bio/icons/svm.png')
        self.classTypeComboBox.addItem(self.svmIcon, 'SVM')
        self.svmPage = QWidget()
        self.paramStack.addWidget(self.svmPage)

        # fields
        self.svmPenaltyLineEdit = QLineEdit('1.0', self.svmPage)
        self.svmKernelComboBox = QComboBox(self.svmPage)
        self.svmKernelComboBox.addItem('rbf')
        self.svmKernelComboBox.addItem('linear')
        self.svmKernelComboBox.addItem('poly')
        self.svmKernelComboBox.addItem('sigmoid')
        self.svmDegreeSpinBox = self.spin_box(3, 1, 5, 1, self.svmPage)
        self.svmDegreeSpinBox.setDisabled(True)
        self.svmGammaLineEdit = QLineEdit('', self.svmPage)
        self.svmCoefLineEdit = QLineEdit('0.0', self.svmPage)
        self.svmClassWeightComboBox = QComboBox(self.svmPage)
        self.svmClassWeightComboBox.addItem('uniform')
        self.svmClassWeightComboBox.addItem('balanced')

        self.svmStopLabel = QLabel('Stopping Criteria:', self.svmPage)
        self.svmTolLabel = QLabel('tol:', self.svmPage)
        self.svmTolLabel.setMinimumWidth(60)
        self.svmTolLineEdit = QLineEdit('1e-3', self.svmPage)
        self.svmMaxIterLabel = QLabel('max_iter:', self.svmPage)
        self.svmMaxIterLabel.setMinimumWidth(60)
        self.svmMaxIterLineEdit = QLineEdit('200', self.svmPage)

        self.svmDoc = QLabel("<a href=\"http://scikit-learn.org/stable/modules/generated/sklearn.svm.SVC.html\">Documentation</a>")
        self.svmDoc.setAlignment(Qt.AlignRight)
        self.svmDoc.setOpenExternalLinks(True)

        # layout
        self.svmLayout = QFormLayout()
        self.svmLayout.addRow('C:', self.svmPenaltyLineEdit)
        self.svmLayout.addRow('kernel:', self.svmKernelComboBox)
        self.svmLayout.addRow('degree:', self.svmDegreeSpinBox)
        self.svmLayout.addRow('gamma:', self.svmGammaLineEdit)
        self.svmLayout.addRow('coef0:', self.svmCoefLineEdit)
        self.svmLayout.addRow('class_weight:', self.svmClassWeightComboBox)

        self.svmSpacer = QSpacerItem(40, 20, QSizePolicy.Minimum, QSizePolicy.Expanding)

        self.svmStopLayout = QHBoxLayout()
        self.svmStopLayout.addWidget(self.svmTolLabel)
        self.svmStopLayout.addWidget(self.svmTolLineEdit)
        self.svmStopLayout.addWidget(self.svmMaxIterLabel)
        self.svmStopLayout.addWidget(self.svmMaxIterLineEdit)

        self.svmPageLayout = QVBoxLayout(self.svmPage)
        self.svmPageLayout.addLayout(self.svmLayout)
        self.svmPageLayout.addItem(self.svmSpacer)
        self.svmPageLayout.addWidget(self.svmStopLabel)
        self.svmPageLayout.addLayout(self.svmStopLayout)

        self.svmPageLayout.addWidget(self.svmDoc)

        ## Naive bayes
        self.nbIcon = QIcon('ml4bio/icons/nb.png')
        self.classTypeComboBox.addItem(self.nbIcon, 'Naive Bayes')
        self.nbPage = QWidget()
        self.paramStack.addWidget(self.nbPage)

        # fields
        self.nbDistributionLabel = QLabel('')
        self.nbAddSmoothDoubleSpinBox = self.double_spin_box(1, 0, 50, 0.5, 1, self.nbPage)
        self.nbFitPriorCheckBox = QCheckBox(self.nbPage)
        self.nbFitPriorCheckBox.setChecked(True)
        self.nbClassPriorLineEdit = QLineEdit('None', self.nbPage)

        self.nbDoc = QLabel("<a href=\"http://scikit-learn.org/stable/modules/generated/sklearn.naive_bayes.GaussianNB.html\">Documentation</a>")
        self.nbDoc.setAlignment(Qt.AlignRight)
        self.nbDoc.setOpenExternalLinks(True)

        # layout
        self.nbLayout = QFormLayout()
        self.nbLayout.addRow('distributon:', self.nbDistributionLabel)
        self.nbLayout.addRow('smoothing:', self.nbAddSmoothDoubleSpinBox)
        self.nbLayout.addRow('fit_prior:', self.nbFitPriorCheckBox)
        self.nbLayout.addRow('class_prior:', self.nbClassPriorLineEdit)

        self.nbSpacer = QSpacerItem(40, 20, QSizePolicy.Minimum, QSizePolicy.Expanding)

        self.nbPageLayout = QVBoxLayout(self.nbPage)
        self.nbPageLayout.addLayout(self.nbLayout)
        self.nbPageLayout.addItem(self.nbSpacer)

        self.nbPageLayout.addWidget(self.nbDoc)

        ### classifier name
        self.classNameLabel = self.title('Classifier Name:', self.modelPage)
        self.classNameLineEdit = QLineEdit(self.modelPage)
        self.classNameLineEdit.setDisabled(True)
        self.classNameLayout = QHBoxLayout()
        self.classNameLayout.addWidget(self.classNameLabel)
        self.classNameLayout.addWidget(self.classNameLineEdit)
        self.modelPageLayout.addLayout(self.classNameLayout)

        ### comment
        self.classCommentLabel = self.title('Comment (no more than 30 characters):', self.modelPage)
        self.classCommentTextEdit = QTextEdit(self.modelPage)
        self.classCommentTextEdit.setMaximumHeight(30)
        self.classCommentTextEdit.setDisabled(True)

        self.modelPageLayout.addWidget(self.classCommentLabel)
        self.modelPageLayout.addWidget(self.classCommentTextEdit)

        ### training status
        self.trainStatusLabel = QLabel()
        self.trainStatusLabel.setAlignment(Qt.AlignRight)
        self.trainStatusLabel.setStyleSheet("color: red")
        self.modelPageLayout.addWidget(self.trainStatusLabel)

        ### reset, stop and train buttons
        self.classResetTrainSpacer = QSpacerItem(40, 20, QSizePolicy.Expanding, QSizePolicy.Minimum)
        self.classResetPushButton = QPushButton('Reset', self.modelPage)
        self.classResetPushButton.setMinimumWidth(90)
        self.classResetPushButton.setDisabled(True)
        self.classTrainPushButton = QPushButton('Train', self.modelPage)
        self.classTrainPushButton.setMinimumWidth(90)
        self.classTrainPushButton.setDefault(True)
        self.classTrainPushButton.setDisabled(True)

        self.classResetTrainLayout = QHBoxLayout()
        self.classResetTrainLayout.addItem(self.classResetTrainSpacer)
        self.classResetTrainLayout.addWidget(self.classResetPushButton)
        self.classResetTrainLayout.addWidget(self.classTrainPushButton)

        self.modelPageLayout.addLayout(self.classResetTrainLayout)

        ### page spacer and line
        self.modelPageSpacer = QSpacerItem(40, 20, QSizePolicy.Minimum, QSizePolicy.Expanding)
        self.modelPageLayout.addItem(self.modelPageSpacer)
        self.modelLine = QFrame(self.modelPage)
        self.modelLine.setFrameShape(QFrame.HLine)
        self.modelLine.setFrameShadow(QFrame.Sunken)
        self.modelPageLayout.addWidget(self.modelLine)

        ### back and next buttons
        self.classBackNextSpacer = QSpacerItem(40, 20, QSizePolicy.Expanding, QSizePolicy.Minimum)
        self.classBackPushButton = QPushButton('Back', self.modelPage)
        self.classBackPushButton.setMinimumWidth(90)
        self.classNextPushButton = QPushButton('Next', self.modelPage)
        self.classNextPushButton.setMinimumWidth(90)
        self.classNextPushButton.setDefault(True)
        self.classNextPushButton.setDisabled(True)
        self.classBackNextLayout = QHBoxLayout()
        self.classBackNextLayout.addItem(self.classBackNextSpacer)
        self.classBackNextLayout.addWidget(self.classBackPushButton)
        self.classBackNextLayout.addWidget(self.classNextPushButton)

        self.modelPageLayout.addLayout(self.classBackNextLayout)

        self.leftPanel.addWidget(self.modelPage)

        # signal handling
        self.classTypeComboBox.currentIndexChanged.connect(self.paramStack.setCurrentIndex)
        self.lrSolverComboBox.currentIndexChanged.connect(self.update_logistic_regression)
        self.nnSolverComboBox.currentIndexChanged.connect(self.update_neural_network)
        self.svmKernelComboBox.currentIndexChanged.connect(self.update_svm)

        self.paramStack.currentChanged.connect(\
            lambda: self.classNameLineEdit.setEnabled(self.paramStack.currentIndex() > 0))
        self.paramStack.currentChanged.connect(\
            lambda: self.classCommentTextEdit.setEnabled(self.paramStack.currentIndex() > 0))
        self.paramStack.currentChanged.connect(\
            lambda: self.classTrainPushButton.setEnabled(self.paramStack.currentIndex() > 0))
        self.paramStack.currentChanged.connect(\
            lambda: self.classResetPushButton.setEnabled(self.paramStack.currentIndex() > 0))

        self.classTrainPushButton.clicked.connect(self.train)
        self.classResetPushButton.clicked.connect(self.reset)

        self.classBackPushButton.clicked.connect(lambda: self.clear(option='train'))
        self.classNextPushButton.clicked.connect(lambda: self.leftPanel.setCurrentIndex(2))

        ########## 3rd page: model selection and testing ##########
        self.testPage = QWidget()
        self.testPageLayout = QVBoxLayout(self.testPage)
        self.testPageTitle = self.page_title('Step 3: Test and Predict', self.testPage)
        self.testPageLayout.addWidget(self.testPageTitle)
        self.testPageLine = QFrame(self.testPage)
        self.testPageLine.setFrameShape(QFrame.HLine)
        self.testPageLine.setFrameShadow(QFrame.Sunken)
        self.testPageLayout.addWidget(self.testPageLine)

        ### classifier selection
        self.classSelectLabel = self.title('Classifier Selection:', self.testPage)
        self.testPageLayout.addWidget(self.classSelectLabel)

        self.classSelectFrame = QFrame(self.testPage)
        self.classSelectFrame.setAutoFillBackground(True)
        self.classSelectLayout = QVBoxLayout(self.classSelectFrame)
        self.bestPerformRadioButton = QRadioButton('Best-Performing', self.classSelectFrame)
        self.bestPerformRadioButton.setChecked(True)
        self.userPickRadioButton = QRadioButton('User-Picked', self.classSelectFrame)
        self.bestPerformSpacer = QSpacerItem(40, 20, QSizePolicy.Expanding, QSizePolicy.Minimum)
        self.metricLabel = QLabel('Metric:', self.classSelectFrame)
        self.metricComboBox = QComboBox(self.classSelectFrame)
        self.metricComboBox.addItem('--- select ---')
        self.metricComboBox.addItem('accuracy')
        self.metricComboBox.addItem('precision')
        self.metricComboBox.addItem('recall')
        self.metricComboBox.addItem('f1')
        self.metricComboBox.addItem('AUROC')
        self.metricComboBox.addItem('AUPRC')
        self.userPickLabel = QLabel('None', self.classSelectFrame)
        self.userPickLabel.setMinimumWidth(150)

        self.bestPerformLayout = QHBoxLayout()
        self.bestPerformLayout.addWidget(self.bestPerformRadioButton)
        self.bestPerformLayout.addItem(self.bestPerformSpacer)
        self.bestPerformLayout.addWidget(self.metricLabel)
        self.bestPerformLayout.addWidget(self.metricComboBox)
        self.classSelectLayout.addLayout(self.bestPerformLayout)

        self.userPickLayout = QHBoxLayout()
        self.userPickLayout.addWidget(self.userPickRadioButton)
        self.userPickLayout.addWidget(self.userPickLabel)
        self.classSelectLayout.addLayout(self.userPickLayout)

        self.testPageLayout.addWidget(self.classSelectFrame)

        ### test button
        self.testPushButton = QPushButton('Test')
        self.testPushButton.setMinimumWidth(90)
        self.testPushButton.setDefault(True)
        self.testPushButton.setDisabled(True)
        self.testSpacer = QSpacerItem(40, 20, QSizePolicy.Expanding, QSizePolicy.Minimum)
        self.testLayout = QHBoxLayout()
        self.testLayout.addItem(self.testSpacer)
        self.testLayout.addWidget(self.testPushButton)
        self.testPageLayout.addLayout(self.testLayout)

        ### prediction button
        self.predictionLabel = self.title('Prediction:', self.testPage)
        self.predictionPushButton = QPushButton('Predict and Save as...', self.testPage)
        self.predictionPushButton.setMaximumWidth(175)
        self.predictionPushButton.setDisabled(True)
        self.predictionLayout = QHBoxLayout()
        self.predictionLayout.addWidget(self.predictionLabel)
        self.predictionLayout.addWidget(self.predictionPushButton)
        self.testPageLayout.addLayout(self.predictionLayout)

        ### page spacer and line
        self.testPageSpacer = QSpacerItem(40, 20,QSizePolicy.Minimum, QSizePolicy.Expanding)
        self.testPageLayout.addItem(self.testPageSpacer)
        self.testLine = QFrame(self.testPage)
        self.testLine.setFrameShape(QFrame.HLine)
        self.testLine.setFrameShadow(QFrame.Sunken)
        self.testPageLayout.addWidget(self.testLine)

        ### back and finish buttons
        self.testBackPushButton = QPushButton('Back', self.testPage)
        self.testBackPushButton.setMinimumWidth(90)
        self.testFinishPushButton = QPushButton('Finish', self.testPage)
        self.testFinishPushButton.setMinimumWidth(90)
        self.testFinishPushButton.setDefault(True)
        self.testFinishPushButton.setDisabled(True)        
        self.testBackFinishSpacer = QSpacerItem(40, 20, QSizePolicy.Expanding, QSizePolicy.Minimum)
        self.testBackFinishLayout = QHBoxLayout()
        self.testBackFinishLayout.addItem(self.testBackFinishSpacer)
        self.testBackFinishLayout.addWidget(self.testBackPushButton)
        self.testBackFinishLayout.addWidget(self.testFinishPushButton)
        self.testPageLayout.addLayout(self.testBackFinishLayout)

        self.leftPanel.addWidget(self.testPage)

        # signal handling
        self.bestPerformRadioButton.toggled.connect(self.metricComboBox.setEnabled)
        self.bestPerformRadioButton.toggled.connect(lambda: self.select(option='best'))
        self.metricComboBox.currentIndexChanged.connect(lambda: self.select(option='best'))

        self.userPickRadioButton.toggled.connect(self.metricComboBox.setDisabled)
        self.userPickRadioButton.toggled.connect(lambda: self.select(option='user'))
        
        self.testPushButton.clicked.connect(self.test)
        self.predictionPushButton.clicked.connect(self.predict)

        self.testBackPushButton.clicked.connect(lambda: self.clear(option='test'))
        self.testFinishPushButton.clicked.connect(self.finish)

        ########## right panel ##########
        self.rightPanelLayout = QVBoxLayout(self.rightPanel)

        self.trainedClassifiersLayout = QHBoxLayout()
        self.trainedClassifiersLabel = self.title('Trained Classifiers:', self.rightPanel)
        self.performanceLabel = QLabel('Show performance on ', self.rightPanel)
        self.performanceComboBox = QComboBox(self.rightPanel)
        self.performanceListView = self.performanceComboBox.view()
        self.performanceComboBox.addItem('validation data')
        self.performanceComboBox.addItem('training data')
        self.performanceComboBox.addItem('test data')
        self.performanceListView.setRowHidden(2, True)

        self.trainedClassifiersSpacer = QSpacerItem(40, 20, QSizePolicy.Expanding, QSizePolicy.Minimum)
        self.trainedClassifiersLayout.addWidget(self.trainedClassifiersLabel)
        self.trainedClassifiersLayout.addItem(self.trainedClassifiersSpacer)
        self.trainedClassifiersLayout.addWidget(self.performanceLabel)
        self.trainedClassifiersLayout.addWidget(self.performanceComboBox)
        self.models_table = QTableWidget(self.rightPanel)
        self.models_table.verticalHeader().setDefaultSectionSize(25)
        self.models_table.verticalHeader().hide()
        self.models_table.setSelectionBehavior(QTableWidget.SelectRows)
        self.models_table.setColumnCount(8)
        self.models_table.setHorizontalHeaderLabels(['Name', 'Type', 'Accuracy', 'Precision', 'Recall', 'F1', 'AUROC', 'AUPRC'])
        self.models_table.setColumnWidth(0, 140)
        self.models_table.setColumnWidth(1, 110)
        self.models_table.setColumnWidth(2, 70)
        self.models_table.setColumnWidth(3, 70)
        self.models_table.setColumnWidth(4, 65)
        self.models_table.setColumnWidth(5, 65)
        self.models_table.setColumnWidth(6, 65)
        self.models_table.setColumnWidth(7, 65)
        self.models_table.setFont(self.font)
        self.table_header = self.models_table.horizontalHeader()
        self.rightPanelLayout.addLayout(self.trainedClassifiersLayout)
        self.rightPanelLayout.addWidget(self.models_table)

        self.visLayout = QHBoxLayout()
        self.visListLayout = QVBoxLayout()
        self.model_summary_ = QTreeWidget(self.rightPanel)
        self.model_summary_.setMaximumWidth(300)
        self.model_summary_.setColumnCount(2)
        self.model_summary_.setHeaderHidden(True)
        self.model_summary_.setColumnWidth(0, 160)
        self.model_summary_.setFont(self.font)
        self.visListLayout.addWidget(self.model_summary_)

        self.visFrame = QFrame(self.rightPanel)
        self.visFrameLayout = QHBoxLayout(self.visFrame)
        self.dataPlotRadioButton = QRadioButton('Data Plot', self.visFrame)
        self.dataPlotRadioButton.setDisabled(True)
        self.rocRadioButton = QRadioButton('ROC', self.visFrame)
        self.confusionMatrixRadioButton = QRadioButton('Confusion Matrix', self.visFrame)
        self.confusionMatrixRadioButton.setChecked(True)
        self.prRadioButton = QRadioButton('Precision-Recall', self.visFrame)
        
        self.visFrameLeftLayout = QVBoxLayout()
        self.visFrameLeftLayout.addWidget(self.dataPlotRadioButton)
        self.visFrameLeftLayout.addWidget(self.rocRadioButton)
        self.visFrameRightLayout = QVBoxLayout()
        self.visFrameRightLayout.addWidget(self.confusionMatrixRadioButton)
        self.visFrameRightLayout.addWidget(self.prRadioButton)
        self.visFrameLayout.addLayout(self.visFrameLeftLayout)
        self.visFrameLayout.addLayout(self.visFrameRightLayout)
        self.visListLayout.addWidget(self.visFrame)
        self.visLayout.addLayout(self.visListLayout)

        self.canvas = FigureCanvas(Figure(figsize=(340, 340)))
        self.canvas.setMaximumWidth(340)
        self.canvas.setMaximumHeight(340)
        self.canvas.setParent(self.rightPanel)
        self.visLayout.addWidget(self.canvas)

        self.rightPanelLayout.addLayout(self.visLayout)

        # signal handling
        self.performanceComboBox.currentIndexChanged.connect(self.switch_metrics)
        self.performanceComboBox.currentIndexChanged.connect(\
            lambda: self.bestPerformRadioButton.setDisabled(self.performanceComboBox.currentIndex() == 2))
        self.performanceComboBox.currentIndexChanged.connect(\
            lambda: self.metricComboBox.setDisabled(self.performanceComboBox.currentIndex() == 2 \
                or not self.bestPerformRadioButton.isChecked()))
        self.performanceComboBox.currentIndexChanged.connect(\
            lambda: self.userPickRadioButton.setDisabled(self.performanceComboBox.currentIndex() == 2))
        self.performanceComboBox.currentIndexChanged.connect(\
            lambda: self.testPushButton.setDisabled(self.performanceComboBox.currentIndex() == 2 \
                or self.selected_model == None))
        self.performanceComboBox.currentIndexChanged.connect(\
            lambda: self.table_header.setSectionsClickable(self.performanceComboBox.currentIndex() != 2))
        self.models_table.cellClicked.connect(self.switch_model)
        self.table_header.sectionClicked.connect(self.sort)

        self.confusionMatrixRadioButton.toggled.connect(self.plot)
        self.rocRadioButton.toggled.connect(self.plot)
        self.prRadioButton.toggled.connect(self.plot)

        # global geometry
        self.resize(1064, 700)
        self.setWindowTitle('ML4Bio')
        self.leftPanel.resize(360, 680)
        self.leftPanel.setStyleSheet("QStackedWidget {background-color:rgb(226, 226, 226)}")
        self.leftPanel.move(10, 10)
        self.rightPanel.resize(680, 680)
        self.rightPanel.move(380, 10)

        self.show()

def main():
    """
    Entry point for console_scripts in setup.py.
    """
    app = QApplication(sys.argv)
    ex = App()
    sys.exit(app.exec_())

if __name__ == '__main__':
    main()



import numpy as np
import pandas as pd
from itertools import cycle
from sklearn import metrics
from PyQt5.QtWidgets import QTreeWidgetItem
from matplotlib.backends.backend_qt5agg import FigureCanvas
from matplotlib.figure import Figure
from matplotlib import pyplot as plt
from matplotlib import rcParams


class ModelMetrics:
	"""
	An instance of this class stores information about a classifier's 
	performance metrics with respect to training, validation or test set.
	Stored metrics include:

		- accuracy
		- precision (average and classwise)
		- recall (average and classwise)
		- f1 score (average and classwise)
		- AUROC (average and classwise)
		- AUPRC (average and classwise)
		- confusion matrix

	:ivar classes\_: valid classes for the classification problem
	:vartype classes\_: list
	:ivar num_classes\_: total number of valid classes
	:vartype num_classes\_: int
	:ivar accuracy\_: accuracy of the classifier
	:vartype accuracy\_: float
	:ivar avg_precision\_: average precision over all classes
	:vartype avg_precision\_: float
	:ivar avg_recall\_: average recall over all classes
	:vartype avg_recall\_: float
	:ivar avg_f1\_: average f1 score over all classes
	:vartype avg_f1\_: float
	:ivar avg_auroc\_: average AUROC over all classes
	:vartype avg_auroc\_: float
	:ivar avg_auprc\_: average AUPRC over all classes
	:vartype avg_auprc\_: float
	:ivar confusion_matrix\_: confusion matrix in percents
	:vartype confusion_matrix\_: 2D array

	The dictionaries below use class names as keys.
	Each value in the dictionaries is a floating-point number.

	:ivar class_precision\_: classwise precisions
	:vartype class_precision\_: dict
	:ivar class_recall\_: classwise recalls
	:vartype class_recall\_: dict
	:ivar class_f1\_: classwise f1 scores
	:vartype class_f1\_: dict
	:ivar class_auroc\_: classwise AUROCs
	:vartype class_auroc\_: dict
	:ivar class_auprc\_: classwise AUPRCs
	:vartype class_auprc\_: dict

	The dictionaries below use class names and 'avg' as keys.
	Each value in the dictionaries is a list of metric values at different decision threshold.

	:ivar fpr\_: FPRs for plotting all ROC curves
	:vartype fpr\_: dict
	:ivar tpr\_: TPRs for plotting all ROC curves
	:vartype tpr\_: dict
	:ivar precision\_: precisions for plotting all PR curves
	:vartype precision\_: dict
	:ivar recall\_: recalls for plotting all PR curves
	:vartype recall\_: dict
	"""

	def __init__(self, classifier, y, y_pred, y_prob, option):
		"""
		Constructs a new instance of the class.

		:param classifier: classifier for evaluation
		:type classifier: sklearn classifier object
		:param y: true labels
		:type y: pandas series
		:param y_pred: predicted labels
		:type y_pred: pandas series
		:param y_prob: classwise probabilities
		:type y_prob: pandas dataframe
		:param option: specification of evaluation method ("holdout", "cv" or "loo")
		:type option: str
		"""
		classes = classifier.classes_
		num_classes = len(classes)
		self.classes_ = classes
		self.num_classes_ = num_classes

		fpr = dict()
		tpr = dict()
		auroc = dict()

		precision = dict()
		recall = dict()
		auprc = dict()

		if option in ['holdout', 'loo']:
			self.accuracy_ = np.around(metrics.accuracy_score(y, y_pred), decimals=3)
			class_precision = metrics.precision_score(y, y_pred, average=None)
			class_recall = metrics.recall_score(y, y_pred, average=None)
			class_f1 = metrics.f1_score(y, y_pred, average=None)
			confusion_matrix = metrics.confusion_matrix(y, y_pred)

			for i in range(0, num_classes):
				fpr[i], tpr[i], _ = metrics.roc_curve(y, y_prob[:, i], pos_label=classes[i])
				auroc[classes[i]] = np.around(metrics.auc(fpr[i], tpr[i]), decimals=3)

				precision[i], recall[i], _ = metrics.precision_recall_curve(y, y_prob[:, i], pos_label=classes[i])
				precision[i] = np.flip(precision[i], 0)
				recall[i] = np.flip(recall[i], 0)
				auprc[classes[i]] = np.around(metrics.auc(recall[i], precision[i]), decimals=3)

		elif option == 'cv':
			k = len(y)
			accuracy = 0
			class_precision = np.zeros(num_classes)
			class_recall = np.zeros(num_classes)
			class_f1 = np.zeros(num_classes)
			confusion_matrix = np.zeros([num_classes, num_classes])

			# classwise metrics w.r.t. each fold
			for y_i, y_pred_i in zip(y, y_pred):
				accuracy += metrics.accuracy_score(y_i, y_pred_i)
				class_precision += metrics.precision_score(y_i, y_pred_i, average=None)
				class_recall += metrics.recall_score(y_i, y_pred_i, average=None)
				class_f1 += metrics.f1_score(y_i, y_pred_i, average=None)
				confusion_matrix += metrics.confusion_matrix(y_i, y_pred_i)

			# classwise metrics average over all folds
			self.accuracy_ = np.around(accuracy / k, decimals=3)
			class_precision = class_precision / k
			class_recall = class_recall / k
			class_f1 = class_f1 / k

			for i in range(0, num_classes):
				fpr_i = dict()
				tpr_i = dict()
				precision_i = dict()
				recall_i = dict()

				for j in range(0, k):
					fpr_i[j], tpr_i[j], _ = metrics.roc_curve(y[j], y_prob[j][:, i], pos_label=classes[i])
					precision_i[j], recall_i[j], _ = metrics.precision_recall_curve(y[j], y_prob[j][:, i], pos_label=classes[i])
				
				fpr[i] = np.unique(np.concatenate([fpr_i[j] for j in range(0, k)]))
				tpr[i] = np.zeros_like(fpr[i])
				recall[i] = np.unique(np.concatenate([recall_i[j] for j in range(0, k)])) 	# all unique x-values
				precision[i] = np.zeros_like(recall[i])

				# interpolate y-values
				for j in range(0, k):
					tpr[i] += np.interp(fpr[i], fpr_i[j], tpr_i[j])
					precision[i] += np.interp(recall[i], np.flip(recall_i[j], 0), np.flip(precision_i[j], 0))
				
				tpr[i] /= k
				precision[i] /= k
				auroc[classes[i]] = np.around(metrics.auc(fpr[i], tpr[i]), decimals=3)
				auprc[classes[i]] = np.around(metrics.auc(recall[i], precision[i]), decimals=3)

		# metrics averaged over all classes
		self.avg_precision_ = np.around(np.mean(class_precision), decimals=3)
		self.avg_recall_ = np.around(np.mean(class_recall), decimals=3)
		self.avg_f1_ = np.around(np.mean(class_f1), decimals=3)

		self.class_precision_ = dict(zip(classes, np.around(class_precision, decimals=3)))
		self.class_recall_ = dict(zip(classes, np.around(class_recall, decimals=3)))
		self.class_f1_ = dict(zip(classes, np.around(class_f1, decimals=3)))

		#confusion_matrix = confusion_matrix / np.sum(confusion_matrix, axis=1)
		self.confusion_matrix_ = np.around(confusion_matrix, decimals=3)
		
		all_fpr = np.unique(np.concatenate([fpr[i] for i in range(0, num_classes)]))
		all_recall = np.unique(np.concatenate([recall[i] for i in range(0, num_classes)]))
		mean_tpr = np.zeros_like(all_fpr)
		mean_precision = np.zeros_like(all_recall)

		for i in range(0, num_classes):
			mean_tpr += np.interp(all_fpr, fpr[i], tpr[i])
			mean_precision += np.interp(all_recall, recall[i], precision[i])

		mean_tpr /= num_classes
		mean_precision /= num_classes

		fpr['avg'] = all_fpr
		tpr['avg'] = mean_tpr
		self.avg_auroc_ = np.around(metrics.auc(fpr['avg'], tpr['avg']), decimals=3)

		precision['avg'] = mean_precision
		recall['avg'] = all_recall
		self.avg_auprc_ = np.around(metrics.auc(recall['avg'], precision['avg']), decimals=3)

		self.fpr_ = fpr
		self.tpr_ = tpr
		self.class_auroc_ = auroc

		self.precision_ = precision
		self.recall_ = recall
		self.class_auprc_ = auprc

	def accuracy(self):
		"""
		Returns accuracy of the classifier.
		"""
		return self.accuracy_

	def precision(self, option='average'):
		"""
		Returns precision of the classifier.

			- option='average': average precision
			- option='classwise': classwise precision

		:param option: 'average' or 'classwise'
		:type option: str
		"""
		if option == 'classwise':	return self.class_precision_
		elif option == 'average':	return self.avg_precision_

	def recall(self, option='average'):
		"""
		Returns recall of the classifier.

			- option='average': average recall
			- option='classwise': classwise recall

		:param option: 'average' or 'classwise'
		:type option: str
		"""
		if option == 'classwise':	return self.class_recall_
		elif option == 'average':	return self.avg_recall_

	def f1(self, option='average'):
		"""
		Returns f1 score of the classifier.

			- option='average': average f1 score
			- option='classwise': classwise f1 score

		:param option: 'average' or 'classwise'
		:type option: str
		"""
		if option == 'classwise':	return self.class_f1_
		elif option == 'average':	return self.avg_f1_

	def auroc(self, option='average'):
		"""
		Returns AUROC of the classifier.

			- option='average': average AUROC
			- option='classwise': classwise AUROC

		:param option: 'average' or 'classwise'
		:type option: str
		"""
		if option == 'classwise':	return self.class_auroc_
		elif option == 'average':	return self.avg_auroc_

	def auprc(self, option='average'):
		"""
		Returns AUPRC of the classifier.

			- option='average': average AUPRC
			- option='classwise': classwise AUPRC

		:param option: 'average' or 'classwise'
		:type option: str
		"""
		if option == 'classwise':	return self.class_auprc_
		elif option == 'average':	return self.avg_auprc_

	def summary(self, parent):
		"""
		Displays a summary of performance metrics.
		The summary is organized in a tree structure.

		:param parent: the parent item in the model summary
		:type parent: QTreeWidgetItem
		"""
		accuracy_item = QTreeWidgetItem(parent)
		accuracy_item.setText(0, 'Accuracy')
		accuracy_item.setText(1, str(self.accuracy()))

		precision_item = QTreeWidgetItem(parent)
		precision_item.setText(0, 'Precision')
		precision_item.setText(1, str(self.precision()))
		class_precision = self.precision('classwise')
		self.class_summary(precision_item, class_precision)

		recall_item = QTreeWidgetItem(parent)
		recall_item.setText(0, 'Recall')
		recall_item.setText(1, str(self.recall()))
		class_recall = self.recall('classwise')
		self.class_summary(recall_item, class_recall)

		f1_item = QTreeWidgetItem(parent)
		f1_item.setText(0, 'F1')
		f1_item.setText(1, str(self.f1()))
		class_f1 = self.f1('classwise')
		self.class_summary(f1_item, class_f1)

		auroc_item = QTreeWidgetItem(parent)
		auroc_item.setText(0, 'AUROC')
		auroc_item.setText(1, str(self.auroc()))
		class_auroc = self.auroc('classwise')
		self.class_summary(auroc_item, class_auroc)

		auprc_item = QTreeWidgetItem(parent)
		auprc_item.setText(0, 'AUPRC')
		auprc_item.setText(1, str(self.auprc()))
		class_auprc = self.auprc('classwise')
		self.class_summary(auprc_item, class_auprc)

	def class_summary(self, parent, class_metric):
		"""
		Returns a summary of metrics with respect of each class.

		:param parent: the parent node in the tree view
		:type parent: QTreeWidgetItem
		:param class_metric: the classwise values of a particular metric
		:type class_metric: dict
		"""
		for c in class_metric:
			class_metric_item = QTreeWidgetItem(parent)
			class_metric_item.setText(0, str(c))
			class_metric_item.setToolTip(0, str(c))
			class_metric_item.setText(1, str(class_metric[c]))

	def plot_ROC(self, canvas):
		"""
		Plots ROC curves. This includes all classwise curves and a macro-averaged curve.

		Reference:
			http://scikit-learn.org/stable/auto_examples/model_selection/plot_roc.html

		:param canvas: the canvas where the plot is drawn
		:type canvas: FigureCanvas
		"""
		fpr = self.fpr_
		tpr = self.tpr_
		auc = self.class_auroc_
		classes = self.classes_
		num_classes = self.num_classes_

		rcParams.update({'font.size': 7})
		canvas.figure.clear()
		ax = canvas.figure.subplots()

		ax.plot(fpr['avg'], tpr['avg'], label='avg (area={0})'.format(self.avg_auroc_), \
			color = 'black', linewidth=2, linestyle='--')

		colors = cycle(['blue', 'orange', 'green', 'red', 'yellow', 'purple', 'cyan'])
		for i, color in zip(range(0, num_classes), colors):
			ax.plot(fpr[i], tpr[i], label='{0} (area={1})'.format(classes[i], auc[classes[i]]), \
				color=color, linewidth=1)

		ax.plot([0 ,1], [0, 1], color='lightgray', linewidth=1, linestyle='--')
		ax.set_xlim([0.0, 1.0])
		ax.set_ylim([0.0, 1.05])
		ax.set_xlabel('FPR')
		ax.set_ylabel('TPR')
		ax.legend(loc='lower right')

		canvas.figure.tight_layout()
		canvas.draw()

	def plot_precision_recall(self, canvas):
		"""
		Plots precision-recall curves. This includes all classwise curves and a macro-averaged curve.

		Reference:
			http://scikit-learn.org/stable/auto_examples/model_selection/plot_precision_recall.html

		:param canvas: the canvas where the plot is drawn
		:type canvas: FigureCanva
		"""
		precision = self.precision_
		recall = self.recall_
		auc = self.class_auprc_
		classes = self.classes_
		num_classes = self.num_classes_

		rcParams.update({'font.size': 7})
		canvas.figure.clear()
		ax = canvas.figure.subplots()

		ax.plot(recall['avg'], precision['avg'], label='avg (area={0})'.format(self.avg_auprc_), \
			color = 'black', linewidth=2, linestyle='--')

		colors = cycle(['blue', 'orange', 'green', 'red', 'yellow', 'purple', 'cyan'])
		for i, color in zip(range(0, num_classes), colors):
			ax.plot(recall[i], precision[i], label='{0} (area={1})'.format(classes[i], auc[classes[i]]), \
				color=color, linewidth=1)

		# plot iso-f1 curves
		f1_scores = np.linspace(0.2, 0.8, num=4)
		x_top = np.array([0.1, 0.25, 0.45, 0.7])

		for i, f1_score in enumerate(f1_scores):
			x = np.linspace(0.01, 1)
			y = f1_score * x / (2 * x - f1_score)
			ax.plot(x[y >= 0], y[y >= 0], color='lightgray', alpha=0.2, linewidth=1)
			ax.annotate('f1={0:.1f}'.format(f1_score), xy=(x_top[i], 1)) 

		ax.set_xlim([0.0, 1.0])
		ax.set_ylim([0.0, 1.05])
		ax.set_xlabel('Recall')
		ax.set_ylabel('Precision')
		ax.legend(loc='lower right')

		canvas.figure.tight_layout()
		canvas.draw()

	def plot_confusion_matrix(self, canvas):
		"""
		Plots confusion matrix. The entries are normalized to lie between 0 and 1.

		Reference:
			http://scikit-learn.org/stable/auto_examples/model_selection/plot_confusion_matrix.html

		:param canvas: the canvas where the plot is drawn
		:type canvas: FigureCanva
		"""
		classes = self.classes_
		num_classes = self.num_classes_
		cm = self.confusion_matrix_

		rcParams.update({'font.size': 7})
		canvas.figure.clear()
		ax = canvas.figure.subplots()
		tick_marks = np.arange(num_classes)

		ax.imshow(cm, interpolation='nearest', cmap=plt.cm.Blues)
		ax.set_xticks(tick_marks)
		ax.set_yticks(tick_marks)
		ax.set_xticklabels(classes, rotation=45)
		ax.set_yticklabels(classes, rotation=45)

		cutoff = cm.min() + (cm.max()-cm.min()) / 2

		for i in range(0, cm.shape[0]):
			for j in range(0, cm.shape[1]):
				ax.text(j, i, cm[i, j], horizontalalignment='center', \
					color='white' if cm[i, j] > cutoff else 'black')

		ax.set_xlabel('Predicted class')
		ax.set_ylabel('True class')

		canvas.figure.tight_layout()
		canvas.draw()


# uses https://github.com/greenelab/manubot/blob/master/setup.py as a template
import pathlib
import re

import setuptools

directory = pathlib.Path(__file__).parent.resolve()

# version
init_path = directory.joinpath('ml4bio', '__init__.py')
text = init_path.read_text()
pattern = re.compile(r"^__version__ = ['\"]([^'\"]*)['\"]", re.MULTILINE)
version = pattern.search(text).group(1)

# long_description
readme_path = directory.joinpath('README.md')
long_description = readme_path.read_text()

setuptools.setup(
    # Package details
    name='ml4bio',
    version=version,
    url='https://github.com/gitter-lab/ml4bio',
    description='A graphical interface for sklearn classification to introduce machine learning to biologists',
    long_description_content_type='text/markdown',
    long_description=long_description,
    license='MIT',

    # Author details
    author='Fangzhou Mu',
    author_email='fmu2@wisc.edu',

    # Package topics
    keywords='education machine-learning classification',
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Science/Research',
        'Topic :: Scientific/Engineering :: Bio-Informatics',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3 :: Only',
        'Programming Language :: Python :: 3.5',
    ],

    packages=setuptools.find_packages(),

    # Specify python version
    python_requires='>=3.5',

    # Run-time dependencies
    install_requires=[
        'matplotlib',
        'numpy',
        'pandas',
        # The package name is PyQt5 on PyPI https://pypi.org/project/PyQt5/
        # pip does not recognize pyqt
        #'pyqt>=5',
        'scikit-learn',
        'scipy',
    ],

    # Additional groups of dependencies
    extras_require={
    },

    # Create command line script
    entry_points={
        'console_scripts': [
            'ml4bio = ml4bio.ml4bio:main',
        ],
    },

    # Include package data files from MANIFEST.in
    include_package_data=True,
)
